#!/usr/bin/env python3
# --------------------( LICENSE                            )--------------------
# Copyright 2014-2025 by Alexis Pietak & Cecil Curry.
# See "LICENSE" for further details.

'''
Low-level utility functions specific to single-frame plots.
'''

# ....................{ IMPORTS                            }....................
import matplotlib.cm as cm
import matplotlib.pyplot as plt
import numpy as np
import numpy.ma as ma
# from betse.util.io.log import logs
from matplotlib.collections import LineCollection, PolyCollection
from scipy import interpolate

# ....................{ PLOTTERS                           }....................
def plotSingleCellVData(sim,celli,p,fig=None,ax=None, lncolor='k'):

    tvect_data=[x[celli]*1000 for x in sim.vm_time]

    if fig is None:
        fig = plt.figure()# define the figure and axes instances
    if ax is None:
        ax = plt.subplot(111)

    ax.plot(sim.time, tvect_data,lncolor,linewidth=2.0)

    if p.GHK_calc is True:
        tvect_data_ghk = [x[p.visual.single_cell_index]*1000 for x in sim.vm_GHK_time]
        ax.plot(sim.time, tvect_data_ghk,'r',linewidth=2.0)

    ax.set_xlabel('Time [s]')
    ax.set_ylabel('Voltage [mV]')

    return fig, ax

def plotSingleCellCData(simdata_time,simtime,ioni,celli,fig=None,ax=None,lncolor='b',ionname='ion'):

    # ccIon_cell = [arr[ioni][celli] for arr in simdata_time]
    ccIon_cell = []

    for carray in simdata_time:
        conc = carray[ioni][celli]
        ccIon_cell.append(conc)

    if fig is None:
        fig = plt.figure()# define the figure and axes instances
    if ax is None:
        ax = plt.subplot(111)
        #ax = plt.axes()

    lab = ionname

    ax.plot(simtime, ccIon_cell,lncolor,label=lab)
    ax.set_xlabel('Time [s]')
    ax.set_ylabel('Concentration [mol/m3]')

    return fig, ax

def plotSingleCellData(simtime,simdata_time,celli,fig=None,ax=None,lncolor='b',lab='Data'):

    data_cell = [arr[celli] for arr in simdata_time]

    if fig is None:
        fig = plt.figure()# define the figure and axes instances
    if ax is None:
        ax = plt.subplot(111)

    xmin = simtime[0]
    xmax = simtime[-1]
    ymin = np.min(data_cell)
    ymax = np.max(data_cell)

    ax.plot(simtime, data_cell,lncolor,label=lab)
    ax.set_xlabel('Time [s]')
    ax.set_ylabel(lab)

    return fig, ax

def plotFFT(simtime,simdata_time,celli,fig=None,ax=None,lncolor='b',lab='Data'):
    """
    Calculates the FFT for time-series data defined on a single cell (p.visual.single_cell_index)
    and returns a plot of the spectrum in frequency-space.

    Parameters
    -----------
    simtime:            The time vector for the plot
    simdata_time:       The full data for the plot define on all cell centres of the cluster at all sample times
    celli:              The single cell index to extract data for
    fig:                A handle to an existing figure (default None; function creates the figure)
    ax:                 Handle to an existing axis (default None; function creates the axis)
    lncolor:            Line colour for the plot
    lab:                Label for the data type (e.g. "Vmem [V]" or "Body Force [N/m3]")

    Returns
    --------
    fig, ax     Handles to the figure and axis of the FFT plot
    """

    if fig is None:
        fig = plt.figure()
    if ax is None:
        ax = plt.subplot(111)

    sample_size = len(simtime)
    sample_spacing = simtime[1] - simtime[0]

    cell_data_o = [arr[celli] for arr in simdata_time]
    # membranes_midpoint_data = ((1/sample_size)*(cell_data_o/np.mean(cell_data_o)) )   # normalize the signal
    cell_data = (1/sample_size)*(cell_data_o - np.mean(cell_data_o))

    f_axis = np.fft.rfftfreq(sample_size, d=sample_spacing)
    fft_data_o = np.fft.rfft(cell_data)
    fft_data = np.sqrt(np.real(fft_data_o)**2 + np.imag(fft_data_o)**2)

    xmin = f_axis[0]
    xmax = f_axis[-1]
    ymin = np.min(fft_data)
    ymax = np.max(fft_data)

    ax.plot(f_axis,fft_data)
    ax.axis([xmin,xmax,ymin,ymax])

    ax.set_xlabel('Frequency [1/s]')
    ax.set_ylabel('Signal Power')

    return fig, ax


def plotHetMem(sim,cells, p, fig=None, ax=None, zdata=None,clrAutoscale = True, clrMin = None, clrMax = None,
    clrmap=None,edgeOverlay = True,pointOverlay=None, number_cells = False, number_mems = False,
    number_ecm = False, current_overlay = False,plotIecm = False):
        """
        This plotting method assigns color-data to each node in the cell cluster that has distinct
        membrane domains for each cell. Data is interpolated to generate a smooth surface plot.
        The method returns a plot instance (fig, axes)

        When using p.is_ecm, this plotting method overrides both plotPolyData and plotCellData.

        Parameters
        ----------
        zdata                  A data array with each scalar entry corresponding to a point in
                               mem_mids_flat. If not specified the default is z=1. If 'random'
                               is specified the method creates random vales from 0 to 1..

        clrmap                 The colormap to use for plotting. Must be specified as cm.mapname. A list of
                               available mapnames is supplied at
                               http://matplotlib.org/examples/color/colormaps_reference.html

        clrAutoscale           If True, the colorbar is autoscaled to the max and min of zdata.

        clrMin                 Sets the colorbar to a user-specified minimum value.

        clrMax                 Set the colorbar to a user-specified maximum value


        edgeOverlay             This option allows the user to specify whether or not they want cell edges overlayed.
                                Default is False, set to True to use.

        pointOverlay            This option allows user to specify whether or not they want cell_centre points plotted
                                Default is False, set to True to use.

        number_cells,           Booleans that control whether or not cell, membrane, and ecm spaces are labeled
        number_ecm,             with their indices.
        number_mems


        Returns
        -------
        fig, ax                Matplotlib figure and axes instances for the plot.

        Notes
        -------
        Uses `matplotlib.pyplot` and `numpy` arrays. With `edgeOverlay` and
        `pointOverlay` equal to `None`, this is computationally fast and *is*
        recommended for plotting data on large collectives.
        """

        if fig is None:
            fig = plt.figure()# define the figure and axes instances
        if ax is None:
            ax = plt.subplot(111)

        if clrmap is None:
            clrmap = p.default_cm

        if zdata is None:
            zdata = np.ones((p.plot_grid_size,p.plot_grid_size))

        ax.axis('equal')

        xmin = cells.xmin*p.um
        xmax = cells.xmax*p.um
        ymin = cells.ymin*p.um
        ymax = cells.ymax*p.um

        ax.axis([xmin,xmax,ymin,ymax])

        if p.plotMask is True:
            zdata = ma.masked_array(zdata, np.logical_not(cells.maskM))

        meshplt = plt.imshow(zdata,origin='lower',extent=[xmin,xmax,ymin,ymax],cmap=clrmap)

        if pointOverlay is True:
            ax.scatter(
                p.um*cells.mem_mids_flat[:,0],
                p.um*cells.mem_mids_flat[:,1], c='k',)

        if edgeOverlay is True:
            # cell_edges_flat, _ , _= tb.flatten(cells.mem_edges)
            cell_edges_flat = p.um*cells.mem_edges_flat
            coll = LineCollection(cell_edges_flat,colors='k')
            coll.set_alpha(0.5)
            ax.add_collection(coll)

        if zdata is not None:
            # Add a colorbar for the mesh plot:
            maxval = round(np.max(1000*sim.vm_time[-1]),1)
            minval = round(np.min(1000*sim.vm_time[-1]),1)
            checkval = maxval - minval

            if checkval == 0:
                minval = minval - 0.1
                maxval = maxval + 0.1

        if zdata is not None and clrAutoscale is True:
            meshplt.set_clim(minval,maxval)
            ax_cb = fig.colorbar(meshplt,ax=ax)

        elif clrAutoscale is False:

            meshplt.set_clim(clrMin,clrMax)
            ax_cb = fig.colorbar(meshplt,ax=ax)

        else:
            ax_cb = None

        if number_cells is True:

            for i,cll in enumerate(cells.cell_centres):
                ax.text(p.um*cll[0],p.um*cll[1],i,ha='center',va='center')

        if number_mems is True:

            for i,mem in enumerate(cells.mem_mids_flat):
                ax.text(p.um*mem[0],p.um*mem[1],i,ha='center',va='center')

        if current_overlay is True:

            I_overlay(sim,cells,p,ax,plotIecm)

        return fig, ax, ax_cb


def plotPolyData(
    sim,
    cells,
    p,
    fig = None,
    ax = None,
    zdata = None,
    clrAutoscale: bool = True,
    clrMin = None,
    clrMax = None,
    clrmap = None,
    number_cells: bool = False,
    current_overlay: bool = False,
    plotIecm: bool =False,
):
    """
    Assigns color-data to each polygon in a cell cluster diagram and returns a
    plot instance (fig, axes).

    Parameters
    ----------
    cells : Cells
        Data structure holding all world information about cell geometry.
    zdata : optional[numpy.ndarray]
        A data array with each scalar entry corresponding to a cell's data
        value (for instance, concentration or voltage). If zdata is not
        supplied, the cells will be plotted with a uniform color; if zdata
        is the string `random`, a random data set will be created and
        plotted.
    clrAutoscale : optional[bool]
        If `True`, the colorbar is autoscaled to the max and min of zdata.
    clrMin : optional[float]
        Set the colorbar to a user-specified minimum value.
    clrMax : optional[float]
        Set the colorbar to a user-specified maximum value.
    clrmap : optional[matplotlib.cm]
        The colormap to use for plotting. Must be specified as cm.mapname.
        A list of available mapnames is supplied at:
        http://matplotlib.org/examples/color/colormaps_reference.html

    Returns
    -------
    fig, ax
        Matplotlib figure and axes instances for the plot.

    Notes
    -------
    This method Uses `matplotlib.collections.PolyCollection`,
    `matplotlib.cm`, `matplotlib.pyplot`, and numpy arrays and hence is
    computationally slow. Avoid calling this method for large collectives
    (e.g., larger than 500 x 500 um).
    """

    if fig is None:
        fig = plt.figure()# define the figure and axes instances
    if ax is None:
        ax = plt.subplot(111)
        #ax = plt.axes()

    if zdata is None:  # if user doesn't supply data
        z = np.ones(len(cells.cell_verts)) # create flat data for plotting

    #FIXME: This is a bit cumbersome. Ideally, a new "is_zdata_random"
    #boolean parameter defaulting to "False" should be tested, instead.
    #Whack-a-mole with a big-fat-pole!

    # If random data is requested, do so. To avoid erroneous and expensive
    # elementwise comparisons when "zdata" is neither None nor a string,
    # "zdata" must be guaranteed to be a string *BEFORE* testing this
    # parameter as a string. Numpy prints scary warnings otherwise: e.g.,
    #
    #     FutureWarning: elementwise comparison failed; returning scalar
    #     instead, but in the future will perform elementwise comparison
    elif isinstance(zdata, str) and zdata == 'random':
        z = np.random.random(len(cells.cell_verts)) # create some random data for plotting
    else:
        z = zdata

    # Make the polygon collection and add it to the plot.
    if clrmap is None:
        #clrmap = p.default_cm
        clrmap = cm.rainbow

    if p.showCells:
        # print(f'????   ax type: {type(ax)}????')
        # print(f'????cells type: {type(cells)}????')
        coll, ax = cell_mosaic(
            data=z, ax=ax, cells=cells, p=p, clrmap=p.default_cm)
    else:
        coll, ax = cell_mesh(
            z, ax, cells, p, p.default_cm)

    # points = np.multiply(cells.cell_verts, p.um)
    #
    # coll = PolyCollection(points, array=z, cmap=clrmap, edgecolors='none')
    # ax.add_collection(coll)
    ax.axis('equal')

    # Add a colorbar for the PolyCollection

    if zdata is not None and clrAutoscale is True:
        maxval = np.max(zdata,axis=0)
        minval = np.min(zdata,axis=0)

        coll.set_clim(minval,maxval)
        ax_cb = fig.colorbar(coll,ax=ax)

    elif clrAutoscale is False and zdata is not None:
        coll.set_clim(clrMin,clrMax)
        ax_cb = fig.colorbar(coll,ax=ax)

    elif zdata is None:
        ax_cb = None

    if number_cells is True:
        for i,cll in enumerate(cells.cell_centres):
            ax.text(p.um*cll[0],p.um*cll[1],i,ha='center',va='center')

    if current_overlay is True:
        streams, ax = I_overlay(sim,cells,p,ax,plotIecm)

    xmin = cells.xmin*p.um
    xmax = cells.xmax*p.um
    ymin = cells.ymin*p.um
    ymax = cells.ymax*p.um

    ax.axis([xmin,xmax,ymin,ymax])

    return fig,ax,ax_cb


def plotPrettyPolyData(data, sim, cells, p, fig=None, ax=None, clrAutoscale = True, clrMin = None, clrMax = None,
    clrmap = None, number_cells=False, current_overlay = False,plotIecm=False):
        """
        Assigns color-data to each polygon mem-mid, vertex and cell centre in a cell cluster
        diagram and returns a plot instance (fig, axes).

        Parameters
        ----------
        cells : Cells
            Data structure holding all world information about cell geometry.
        data : [numpy.ndarray]
            A data array with each scalar entry corresponding to a cell's data
            value (for instance, concentration or voltage) at cell membranes. If zdata is not
            supplied, the cells will be plotted with a uniform color; if zdata
            is the string `random`, a random data set will be created and
            plotted.
        clrAutoscale : optional[bool]
            If `True`, the colorbar is autoscaled to the max and min of zdata.
        clrMin : optional[float]
            Set the colorbar to a user-specified minimum value.
        clrMax : optional[float]
            Set the colorbar to a user-specified maximum value.
        clrmap : optional[matplotlib.cm]
            The colormap to use for plotting. Must be specified as cm.mapname.
            A list of available mapnames is supplied at:
            http://matplotlib.org/examples/color/colormaps_reference.html

        Returns
        -------
        fig, ax
            Matplotlib figure and axes instances for the plot.

        Notes
        -------
        This method Uses `matplotlib.collections.PolyCollection`,
        `matplotlib.cm`, `matplotlib.pyplot`, and numpy arrays and hence is
        computationally slow. Avoid calling this method for large collectives
        (e.g., larger than 500 x 500 um).
        """

        if fig is None:
            fig = plt.figure()# define the figure and axes instances
        if ax is None:
            ax = plt.subplot(111)

        # data processing -- map to verts:
        data_verts = np.dot(data, cells.matrixMap2Verts)

        # define colorbar limits for the PolyCollection

        if clrAutoscale is True:
            maxval = data_verts.max()
            minval = data_verts.min()
            # maxval = data_verts.max()
            # minval = data_verts.min()

        else:
            maxval = clrMax
            minval = clrMin


        # Make the polygon collection and add it to the plot.
        if clrmap is None:
            clrmap = p.default_cm

        if p.showCells is True:
            coll, ax = pretty_patch_plot(
                data_verts,ax,cells,p,p.default_cm, cmin=minval, cmax=maxval)
        else:
            coll, ax = cell_mesh(data,ax,cells,p,p.default_cm)

        # add a colorbar
        coll.set_clim(minval, maxval)
        ax_cb = fig.colorbar(coll, ax=ax)

        ax.axis('equal')

        if number_cells is True:
            for i,cll in enumerate(cells.cell_centres):
                ax.text(p.um*cll[0],p.um*cll[1],i,ha='center',va='center')

        if current_overlay is True:
            streams, ax = I_overlay(sim,cells,p,ax,plotIecm)

        xmin = cells.xmin*p.um
        xmax = cells.xmax*p.um
        ymin = cells.ymin*p.um
        ymax = cells.ymax*p.um

        ax.axis([xmin,xmax,ymin,ymax])

        return fig,ax,ax_cb

def plotVectField(Fx,Fy,cells,p,plot_ecm = False,title = 'Vector field',cb_title = 'Field [V/m]',
                    colorAutoscale = True, minColor = None, maxColor=None):

    fig = plt.figure()
    ax = plt.subplot(111)

    if plot_ecm is True:

        efield = np.sqrt(Fx**2 + Fy**2)

        msh = ax.imshow(efield,origin='lower', extent = [cells.xmin*p.um, cells.xmax*p.um, cells.ymin*p.um,
            cells.ymax*p.um],cmap=p.background_cm)

        vplot, ax = env_quiver(Fx,Fy,ax,cells,p)

        tit_extra = 'Extracellular'

    elif plot_ecm is False:

        efield = np.sqrt(Fx**2 + Fy**2)

        msh, ax = cell_mesh(efield,ax,cells,p,p.background_cm)

        vplot, ax = cell_quiver(Fx,Fy,ax,cells,p)

        tit_extra = 'Intracellular'

    ax.axis('equal')

    xmin = cells.xmin*p.um
    xmax = cells.xmax*p.um
    ymin = cells.ymin*p.um
    ymax = cells.ymax*p.um

    ax.axis([xmin,xmax,ymin,ymax])

    if colorAutoscale is False:
        msh.set_clim(minColor,maxColor)

    cb = fig.colorbar(msh)

    tit = title
    ax.set_title(tit)
    ax.set_xlabel('Spatial distance [um]')
    ax.set_ylabel('Spatial distance [um]')
    cb.set_label(cb_title)

    return fig, ax, cb


def plotStreamField(
    Fx,Fy,
    cells,
    p,
    plot_ecm: bool = False,
    title: str = 'Vector field',
    cb_title: str = 'Field [V/m]',
    show_cells: bool = False,
    colorAutoscale: bool = True,
    minColor = None,
    maxColor = None,
):

    fig = plt.figure()
    ax = plt.subplot(111)

    if plot_ecm:
        efield = np.sqrt(Fx**2 + Fy**2)
        # msh = ax.imshow(
        #     efield,
        #     origin='lower',
        #     extent=[cells.xmin*p.um, cells.xmax*p.um, cells.ymin*p.um, cells.ymax*p.um],
        #     cmap=p.background_cm,
        # )
        splot, ax = env_stream(Fx, Fy, ax, cells, p, cmap=p.background_cm)
        tit_extra = 'Extracellular'
    else:
        efield = np.sqrt(Fx**2 + Fy**2)

        # msh, ax = cell_mesh(efield,ax,cells,p,p.background_cm)
        splot, ax = cell_stream(
            Fx, Fy,
            ax, cells, p,
            show_cells=show_cells,
            cmap=p.background_cm,
        )
        tit_extra = 'Intracellular'

    ax.axis('equal')

    xmin = cells.xmin*p.um
    xmax = cells.xmax*p.um
    ymin = cells.ymin*p.um
    ymax = cells.ymax*p.um

    ax.axis([xmin, xmax, ymin, ymax])

    if not colorAutoscale:
        splot.lines.set_clim(minColor, maxColor)

    cb = fig.colorbar(splot.lines)

    ax.set_title(title)
    ax.set_xlabel('Spatial distance [um]')
    ax.set_ylabel('Spatial distance [um]')
    cb.set_label(cb_title)

    return fig, ax, cb


def plotMemData(cells, p, fig= None, ax = None, zdata=None,clrmap=None):
        """

        Assigns color-data to edges in a 2D Voronoi diagram and returns a plot instance (fig, axes)

        Parameters
        ----------
        zdata_t                  A data array with each scalar entry corresponding to a polygon entry in
                               vor_verts. If not specified the default is z=1. If 'random'
                               is specified the method creates random vales from 0 to 1..

        clrmap                 The colormap to use for plotting. Must be specified as cm.mapname. A list of
                               available mapnames is supplied at
                               http://matplotlib.org/examples/color/colormaps_reference.html
                               Default is cm.rainbow. Good options are cm.coolwarm, cm.Blues, cm.jet


        Returns
        -------
        fig, ax                Matplotlib figure and axes instances for the plot.

        Notes
        -------
        Uses matplotlib.collections LineCollection, matplotlib.cm, matplotlib.pyplot and numpy arrays
        Computationally slow -- not recommended for large collectives (500 x 500 um max)

        """

        if fig is None:
            fig = plt.figure()# define the figure and axes instances
        if ax is None:
            ax = plt.subplot(111)

        cell_edges_flat = p.um*cells.mem_edges_flat

        if zdata is None:
            z = np.ones(len(cell_edges_flat))
        #FIXME: This is a bit cumbersome. Ideally, a new "is_zdata_random"
        #boolean parameter defaulting to "False" should be tested, instead.
        #Whack-a-mole with a big-fat-pole!

        # If random data is requested, do so. To avoid erroneous and expensive
        # elementwise comparisons when "zdata" is neither None nor a string,
        # "zdata" must be guaranteed to be a string *BEFORE* testing this
        # parameter as a string. Numpy prints scary warnings otherwise: e.g.,
        #
        #     FutureWarning: elementwise comparison failed; returning scalar
        #     instead, but in the future will perform elementwise comparison
        elif isinstance(zdata, str) and zdata == 'random':
            z = np.random.random(len(cell_edges_flat))
        else:
            z = zdata

        if clrmap is None:
            clrmap = cm.rainbow

        coll = LineCollection(cell_edges_flat, array=z, cmap=clrmap,linewidths=4.0)
        ax.add_collection(coll)

        # coll.set_clim(0,3)

        ax.axis('equal')

        # Add a colorbar for the Line Collection
        if zdata is not None:
            ax_cb = fig.colorbar(coll, ax=ax)

        ax.axis('equal')

        xmin = cells.xmin*p.um
        xmax = cells.xmax*p.um
        ymin = cells.ymin*p.um
        ymax = cells.ymax*p.um

        ax.axis([xmin,xmax,ymin,ymax])

        return fig, ax, ax_cb

def plotConnectionData(cells, p, fig = None, ax=None, zdata=None,clrmap=None,colorbar = None, pickable=None):
        """
        Assigns color-data to connections between a cell and its nearest neighbours and returns plot instance

        Parameters
        ----------

        zdata_t                  A data array with each scalar entry corresponding to a polygon entry in
                               vor_verts. If not specified the default is z=1. If 'random'
                               is specified the method creates random vales from 0 to 1..

        clrmap                 The colormap to use for plotting. Must be specified as cm.mapname. A list of
                               available mapnames is supplied at
                               http://matplotlib.org/examples/color/colormaps_reference.html
                               Default is cm.rainbow. Good options are cm.coolwarm, cm.Blues, cm.jet


        Returns
        -------
        fig, ax                Matplotlib figure and axes instances for the plot.

        Notes
        -------
        Uses matplotlib.collections LineCollection, matplotlib.cm, matplotlib.pyplot and numpy arrays

        """
        if fig is None:
            fig = plt.figure()# define the figure and axes instances
        if ax is None:
            ax = plt.subplot(111)
            #ax = plt.axes()

        if zdata is None:
            z = np.ones(len(cells.gap_jun_i))
        #FIXME: This is a bit cumbersome. Ideally, a new "is_zdata_random"
        #boolean parameter defaulting to "False" should be tested, instead.
        #Whack-a-mole with a big-fat-pole!

        # If random data is requested, do so. To avoid erroneous and expensive
        # elementwise comparisons when "zdata" is neither None nor a string,
        # "zdata" must be guaranteed to be a string *BEFORE* testing this
        # parameter as a string. Numpy prints scary warnings otherwise: e.g.,
        #
        #     FutureWarning: elementwise comparison failed; returning scalar
        #     instead, but in the future will perform elementwise comparison
        elif isinstance(zdata, str) and zdata == 'random':
            z = np.random.random(len(cells.gap_jun_i))

        else:
            z = zdata

        if clrmap is None:
            clrmap = cm.bone_r  # default colormap

         # Make a line collection and add it to the plot.

        con_segs = cells.cell_centres[cells.gap_jun_i]

        connects = p.um*np.asarray(con_segs)

        coll = LineCollection(connects, array=z, cmap=clrmap, linewidths=4.0, zorder=0)
        coll.set_clim(vmin=0.0,vmax=1.0)
        coll.set_picker(pickable)
        ax.add_collection(coll)

        ax.axis('equal')

        # Add a colorbar for the Line Collection
        if zdata is not None and colorbar == 1:
            ax_cb = fig.colorbar(coll, ax=ax)
        else:
            ax_cb = None

        xmin = cells.xmin*p.um
        xmax = cells.xmax*p.um
        ymin = cells.ymin*p.um
        ymax = cells.ymax*p.um

        ax.axis([xmin,xmax,ymin,ymax])

        return fig, ax, ax_cb


def plotBoundCells(points_flat,bflags,cells, p, fig=None, ax=None):
        """
        Plot elements tagged on the boundary as red points.

        Parameters
        ----------
        points_flat          A flat array of points corresponding to the bflags data structure

        bflags          A nested array of boolean flags indicating boundary tagging

        Returns
        -------
        fig, ax         Matplotlib plotting objects

        Note
        ------
        This particular plot is extremely slow -- intended for cross-checking purposes only!

        """
        if fig is None:
            fig = plt.figure()# define the figure and axes instances
        if ax is None:
            ax = plt.subplot(111)
            #ax = plt.axes()

        points_flat = np.asarray(points_flat)
        bflags = np.asarray(bflags)

        bpoints = points_flat[bflags]

        ax.plot(p.um*points_flat[:,0],p.um*points_flat[:,1],'k.')

        ax.plot(p.um*bpoints[:,0],p.um*bpoints[:,1],'r.')

        # cell_edges_flat, _ , _= tb.flatten(cells.mem_edges)
        cell_edges_flat = p.um*cells.mem_edges_flat
        coll = LineCollection(cell_edges_flat,colors='k')
        coll.set_alpha(0.5)
        ax.add_collection(coll)

        ax.axis('equal')

        xmin = cells.xmin*p.um
        xmax = cells.xmax*p.um
        ymin = cells.ymin*p.um
        ymax = cells.ymax*p.um

        ax.axis([xmin,xmax,ymin,ymax])

        return fig, ax

def plotVects(cells, p, fig=None, ax=None):
        """
        This function plots all unit vectors in the tissue system as a cross-check.
        Normals to cell membranes are shown as red arrows.
        Tangents to cell membranes are black arrows.
        Tangents to ecm edges are shown as green arrows.
        Cell membrane edges are drawn as blue lines.

        To plot streamline and vector plots with data use the pyplot quiver and streamplot functions, respectively.

        """

        if fig is None:
            fig = plt.figure()# define the figure and axes instances

        if ax is None:
            ax = plt.subplot(111)
            #ax = plt.axes()

        s = p.um

        ax.quiver(s*cells.mem_vects_flat[:,0],s*cells.mem_vects_flat[:,1],s*cells.mem_vects_flat[:,4],
                  s*cells.mem_vects_flat[:,5],color='b',label='mem tang')
        ax.quiver(s*cells.mem_vects_flat[:,0],s*cells.mem_vects_flat[:,1],s*cells.mem_vects_flat[:,2],
                  s*cells.mem_vects_flat[:,3],color='g',label ='mem norm')
        # ax.quiver(s*cells.ecm_vects[:,0],s*cells.ecm_vects[:,1],s*cells.ecm_vects[:,2],s*cells.ecm_vects[:,3],color='r')

        # cell_edges_flat, _ , _= tb.flatten(cells.mem_edges)
        cell_edges_flat = p.um*cells.mem_edges_flat
        coll = LineCollection(cell_edges_flat,colors='k')
        ax.add_collection(coll)

        ax.axis('equal')

        xmin = cells.xmin*p.um
        xmax = cells.xmax*p.um
        ymin = cells.ymin*p.um
        ymax = cells.ymax*p.um

        ax.axis([xmin,xmax,ymin,ymax])
        plt.legend()

        return fig, ax

def streamingCurrent(
    sim, cells, p,
    fig=None,
    ax=None,
    plot_Iecm=True,
    zdata=None,
    clrAutoscale=True,
    clrMin=None,
    clrMax=None,
    clrmap=cm.coolwarm,
    edgeOverlay=True,
    number_cells=False,
):

    # Define the figure and axes instances if needed.
    if fig is None:
        fig = plt.figure()
    if ax is None:
        ax = plt.subplot(111)

    ax.axis('equal')

    xmin = cells.xmin*p.um
    xmax = cells.xmax*p.um
    ymin = cells.ymin*p.um
    ymax = cells.ymax*p.um

    ax.axis([xmin,xmax,ymin,ymax])

    #FIXME: There's a fair amount of overlap between the following branches.
    #Treetops swaying in the contumely breeze!
    if p.is_ecm is False or plot_Iecm is False:

        # multiply by 100 to get units of uA/m2
        Jmag_M = 100*np.sqrt(
            sim.I_gj_x_time[-1]**2 + sim.I_gj_y_time[-1]**2) + 1e-30

        J_x = sim.I_gj_x_time[-1]/Jmag_M
        J_y = sim.I_gj_y_time[-1]/Jmag_M

        meshplot = plt.imshow(
            Jmag_M,
            origin='lower',
            extent=[xmin,xmax,ymin,ymax],
            cmap=clrmap,
        )

        ax.streamplot(
            cells.X*p.um, cells.Y*p.um, J_x, J_y,
            density=p.stream_density,
            linewidth=(3.0*Jmag_M/Jmag_M.max()) + 0.5,
            color='k',
            cmap=clrmap,
            # arrowsize=5.0,
        )

        ax.set_title('Final gap junction current density')

    elif plot_Iecm is True:
        # multiply by 100 to get units of uA/m2
        Jmag_M = 100*np.sqrt(
            sim.I_tot_x_time[-1]**2 + sim.I_tot_y_time[-1]**2) + 1e-30

        J_x = sim.I_tot_x_time[-1]/Jmag_M
        J_y = sim.I_tot_y_time[-1]/Jmag_M

        meshplot = plt.imshow(
            Jmag_M,
            origin='lower',
            extent=[xmin,xmax,ymin,ymax],
            cmap=clrmap,
        )

        ax.streamplot(
            cells.X*p.um, cells.Y*p.um, J_x, J_y,
            density=p.stream_density,
            linewidth=(3.0*Jmag_M/Jmag_M.max()) + 0.5,
            color='k',
            cmap=clrmap,
            # arrowsize=5.0,
        )

        ax.set_title('Final total currents')

    if clrAutoscale is True:
        ax_cb = fig.colorbar(meshplot,ax=ax)

    elif clrAutoscale is False:
        meshplot.set_clim(clrMin,clrMax)
        ax_cb = fig.colorbar(meshplot,ax=ax)

    # if p.showCells is True:
    #     # cell_edges_flat, _ , _= tb.flatten(cells.mem_edges)
    #     cell_edges_flat = p.um*cells.mem_edges_flat
    #     coll = LineCollection(cell_edges_flat,colors='k')
    #     coll.set_alpha(0.2)
    #     ax.add_collection(coll)

    if number_cells is True:

        for i,cll in enumerate(cells.cell_centres):
            ax.text(p.um*cll[0],p.um*cll[1],i,ha='center',va='center')

    return fig,ax,ax_cb


def I_overlay(sim,cells,p,ax,plotIecm = False):
    """
    Plots an overlay of simulated currents on an existing plot.

    Parameters
    -----------

    sim         Instance of sim module
    cells       Instance of cells module
    p           Instance of parameters module
    ax          Existing figure axis to plot currents on
    plotIecm    Plot total currents (True) or only those of gap junctions (False)

    Returns
    --------
    streams             Container for streamline plot
    ax                  Modified axis
    """

    if p.is_ecm is False or plotIecm is False:

        Ix = sim.I_cell_x_time[-1]
        Iy = sim.I_cell_y_time[-1]

        streams, ax = cell_stream(Ix, Iy,ax,cells,p)

        ax.set_title('(Intracellular current overlay)')

    elif plotIecm is True:

        Ix = sim.I_tot_x_time[-1]
        Iy = sim.I_tot_y_time[-1]

        streams, ax = env_stream(Ix, Iy,ax,cells,p)

        ax.set_title('(Environment current overlay)')

    return streams, ax


def cell_ave(cells,vm_at_mem):

    """
    Averages Vmem over membrane domains to return a mean value for each cell

    Parameters
    ----------
    cells               An instance of the Cells module cells object
    vm_at_mem           Vmem at individual membrane domains


    Returns
    --------
    v_cell              Cell Vm averaged over the whole cell

    """

    v_cell = []

    for i in cells.cell_i:
        cellinds = (cells.mem_to_cells == i).nonzero()
        v_cell_array = vm_at_mem[cellinds]
        v_cell_ave = np.mean(v_cell_array)
        v_cell.append(v_cell_ave)

    v_cell = np.asarray(v_cell)

    return v_cell


def cell_quiver(datax, datay, ax, cells, p):
    """
    Sets up a vector plot for cell-specific data on an existing axis.

    Parameters
    -----------
    datax, datay    Data defined on cell centres or membrane midpoints
    cells           Instance of cells module
    p               Instance of parameters module
    ax              Existing figure axis to plot currents on

    Returns
    --------
    vplot               Container for vector plot, plotted at cell centres
    ax                  Modified axis
    """

    if len(datax) == len(cells.mem_i):
        Fx = np.dot(cells.M_sum_mems,datax)/cells.num_mems
        Fy = np.dot(cells.M_sum_mems,datay)/cells.num_mems
    else:
        Fx = datax
        Fy = datay

    Fmag = np.sqrt(Fx**2 + Fy**2)

    # normalize the data:
    Fmag[Fmag == 0.0] = 1.0
    Fx = Fx/Fmag
    Fy = Fy/Fmag

    vplot = ax.quiver(
        p.um*cells.cell_centres[:,0],p.um*cells.cell_centres[:,1],Fx,Fy,
        pivot='mid', color=p.vcolor, units='x',
        headwidth=5, headlength=7, zorder=10)

    return vplot, ax


def env_quiver(datax,datay,ax,cells,p):
    """
    Sets up a vector plot for environmental data on an existing axis.

    Parameters
    -----------

    datax, datay    Data defined on environmental grid
    cells           Instance of cells module
    p               Instance of parameters module
    ax              Existing figure axis to plot currents on

    Returns
    --------
    vplot               Container for vector plot, plotted at environmental grid points
    ax                  Modified axis

    """
    F_mag = np.sqrt(datax**2 + datay**2)

    if F_mag.max() != 0.0:
        Fx = datax/F_mag.max()
        Fy = datay/F_mag.max()

    else:
        Fx = datax/F_mag.mean()
        Fy = datay/F_mag.mean()

    vplot = ax.quiver(p.um*cells.xypts[:,0], p.um*cells.xypts[:,1], Fx.ravel(),
        Fy.ravel(), pivot='mid',color = p.vcolor, units='x',headwidth=5, headlength = 7,zorder=10)

    return vplot, ax


def cell_stream(
    datax, datay, ax, cells, p, show_cells: bool = False, cmap = None):
    '''
    Add a streamline plot for cell-specific data to the passed axes.

    Parameters
    -----------
    datax, datay    Data defined on cell centres or membrane midpoints
    ax              Existing figure axes to plot currents on
    cells           Instance of cells module
    p               Instance of parameters module

    Returns
    --------
    streams             Container for stream plot, plotted at plot grid
    ax                  Modified axis
    '''

    if show_cells:
        cell_edges_flat = p.um*cells.mem_edges_flat
        coll = LineCollection(cell_edges_flat,colors='k')
        coll.set_alpha(0.3)
        ax.add_collection(coll)

    if datax.shape != cells.X.shape: # if the data hasn't been interpolated yet...
        Fx = interpolate.griddata(
            (cells.cell_centres[:,0], cells.cell_centres[:,1]),
            datax,
            (cells.X, cells.Y),
            fill_value=0,
            method=p.interp_type,
        )
        Fy = interpolate.griddata(
            (cells.cell_centres[:,0], cells.cell_centres[:,1]),
            datay,
            (cells.X, cells.Y),
            fill_value=0,
            method=p.interp_type,
        )

        Fx = Fx*cells.maskECM
        Fy = Fy*cells.maskECM

    else:
        Fx = datax
        Fy = datay

    # Magnitude of the passed vector field.
    Fmag = np.sqrt(Fx**2 + Fy**2)

    # Substitute all zero magnitudes by 1.0, enabling division by these
    # magnitudes without concern for division-by-zero exceptions.
    Fmag[Fmag == 0.0] = 1.0

    # Normalize this vector field to a unit vector field.
    Fx = Fx/Fmag
    Fy = Fy/Fmag

    # Streamline width.
    line_width = (3.0*Fmag/Fmag.max()) + 0.5

    # Color(s) of each streamline, either as a scalar *OR* an array of the same
    # shape as the "Fx" and "Fy" arrays.
    if cmap is None:
        stream_color = p.vcolor
    else:
        stream_color = Fmag

    streams = ax.streamplot(
        cells.X*p.um,
        cells.Y*p.um,
        Fx, Fy,
        density=p.stream_density,
        linewidth=line_width,
        color=stream_color,
        cmap=cmap,
    )

    return streams, ax


def mem_quiver(datax,datay,ax,cells,p, cmap=None):
    """
    Sets up a quiver plot for membrane vectors with tail at cell centres.

    Parameters
    -----------

    datax, datay    Data vectors defined on membrane midpoints
    cells           Instance of cells module
    p               Instance of parameters module
    ax              Existing figure axis to plot currents on

    Returns
    --------
    mvects             Container for quiver plot
    ax                  Modified axis

    """
    # scaleval = (p.um*cells.R.mean())*7.0
    scaleval = p.um*p.wsx*0.8

    mvects = ax.quiver(
        cells.cell_centres[:, 0][cells.mem_to_cells]*p.um,
        cells.cell_centres[:, 1][cells.mem_to_cells]*p.um,
        datax*cells.R[cells.mem_to_cells]*p.um,
        datay*cells.R[cells.mem_to_cells]*p.um,
        scale=scaleval,
        color=p.vcolor,
        cmap = cmap
    )

    ax.axis('equal')
    ax.axis([cells.xmin*p.um, cells.xmax*p.um, cells.ymin*p.um, cells.ymax*p.um])

    return mvects, ax


def env_stream(datax,datay,ax,cells,p, cmap=None):
    """
    Sets up a streamline plot for environmental data on an existing axis.

    Parameters
    -----------

    datax, datay    Data defined on environmental grid
    cells           Instance of cells module
    p               Instance of parameters module
    ax              Existing figure axis to plot currents on

    Returns
    --------
    streams             Container for stream plot
    ax                  Modified axis

    """

    Fmag = np.sqrt(datax**2 + datay**2) + 1e-30

    if Fmag.all() != 0.0:
        Fx = datax/Fmag
        Fy = datay/Fmag

    if Fmag.max() != 0.0:

        lw = (3.0*Fmag/Fmag.max()) + 0.5

    else:
        lw = 3.0

    # if datax.shape == cells.X.shape:

    streams = ax.streamplot(cells.X*p.um,cells.Y*p.um, Fx, Fy, density=p.stream_density,
            linewidth=lw,color=Fmag, cmap=cmap)

    # elif datax.shape == cells.X.shape:
    #
    #     streams = ax.streamplot(cells.X*p.um,cells.Y*p.um, Fx, Fy,density=p.stream_density,
    #         linewidth=lw,color=p.vcolor,arrowsize=1.5)

    # else:
    #     raise BetseFunctionException("Data input to env_streams function must be \n shaped as cells.X or cells.Xgrid.")

    return streams, ax


def cell_mesh(data, ax, cells, p, clrmap):

    # If the data is defined on membrane midpoints get membrane midpoint coordinates
    if len(data) == len(cells.mem_i):
        # data = np.dot(cells.M_sum_mems,data)/cells.num_mems
        xi = p.um*cells.mem_mids_flat[:,0]
        yi = p.um*cells.mem_mids_flat[:,1]

    elif len(data) == len(cells.cell_i): # otherwise

        xi = p.um*cells.cell_centres[:,0]
        yi = p.um*cells.cell_centres[:,1]


    # data_grid = np.zeros(len(cells.voronoi_centres))
    # data_grid[cells.cell_to_grid] = data

    msh = ax.tripcolor(
        xi,
        yi,
        data,
        shading='gouraud',
        cmap=clrmap,
    )

    return msh, ax

#FIXME: Obsoleted by the more general-purpose, reliable, and efficient
#"betse.science.plot.layer.vector.lyrvecabrupt.LayerCellsVectorAbruptMembranes"
#class. Replace all remaining calls to this function with usage of that class;
#then, remove this function.
def pretty_patch_plot(
    data_verts, ax, cells, p, clrmap,
    cmin=None,
    cmax=None,
    use_other_verts=None
):
    """
    Maps data on mem midpoints to vertices, and
    uses tripcolor on every cell patch to create a
    lovely gradient. Slow but beautiful!

    data:   mem midpoint data for plotting (e.g vm)
    ax:     plot axis
    cells:  cells object
    p:      parameters object
    clrmap: colormap
    cmin, cmax   clim values for the data's colormap

    """


    # data_verts = data

    # colormap clim
    if cmin is None:
        amin = data_verts.min()
        amax = data_verts.max()
    else:
        amin = cmin
        amax = cmax

    # amin = amin + 0.1 * np.abs(amin)
    # amax = amax - 0.1 * np.abs(amax)

    # collection of cell patchs at vertices:
    if use_other_verts is None:
        cell_faces = np.multiply(cells.cell_verts, p.um)
    else:
        cell_faces = np.multiply(use_other_verts, p.um)

    # Cell membrane (Vmem) plotter (slow but beautiful!)
    for i in range(len(cell_faces)):
        x = cell_faces[i][:, 0]
        y = cell_faces[i][:, 1]

        # Average color value of each cell membrane, situated at the midpoint
        # of that membrane. This parameter is referred to as "C" in both the
        # documentation and implementation of the tripcolor() function.
        dati = data_verts[cells.cell_to_mems[i]]

        # "matplotlib.collections.TriMesh" instance providing the
        # Gouraud-shaded triangulation mesh for the non-triangular vertices of
        # this cell from the Delaunay hull of these vertices.
        col_cell = ax.tripcolor(x, y, dati, shading='gouraud', cmap=clrmap)

        #FIXME: No need to manually call set_clim() here. Instead, pass the
        #"vmin=amin, vmax=amax" parameters to the above tripcolor() call.

        # Autoscale this mesh's colours as desired.
        col_cell.set_clim(amin, amax)

    return col_cell, ax


def env_mesh(data, ax, cells, p, clrmap, ignore_showCells=False):
    """
    Sets up a mesh plot for environmental data on an existing axis.

    Parameters
    -----------

    data            Data defined on environmental grid
    cells           Instance of cells module
    p               Instance of parameters module
    ax              Existing figure axis to plot currents on

    Returns
    --------
    mesh_plot           Container for mesh plot
    ax                  Modified axis

    """

    # if p.plotMask is True:
    #     data = ma.masked_array(data, np.logical_not(cells.maskM))

    mesh_plot = ax.imshow(data,origin='lower',
                extent=[p.um*cells.xmin,p.um*cells.xmax,p.um*cells.ymin,p.um*cells.ymax],cmap=clrmap)

    if p.showCells is True and ignore_showCells is False:
        cell_edges_flat = p.um*cells.mem_edges_flat
        coll = LineCollection(cell_edges_flat,colors='k')
        coll.set_alpha(0.5)
        ax.add_collection(coll)

    return mesh_plot, ax


def cell_mosaic(
    data,
    #FIXME: This matplotlib class is dynamically unimportable by @beartype for
    #some reason -- probably a circular import dependency. *shrug*
    # ax: 'matplotlib.axes._axes.Axes',
    ax,
    cells: 'betse.science.cells.Cells',
    p: 'betse.science.parameters.Parameters',
    clrmap: 'matplotlib.colors.Colormap',
):
#FIXME: This matplotlib class is dynamically unimportable by @beartype for
#some reason. *shrug*
# ) -> (PolyCollection, 'matplotlib.axes.Axes'):
    """
    Sets up a mosaic plot for cell data on an existing axis.

    Parameters
    -----------
    data            Data defined on environmental grid
    cells           Instance of cells module
    p               Instance of parameters module
    ax              Existing figure axis to plot currents on

    Returns
    --------
    collection          Container for mosaic plot
    ax                  Modified axis
    """

    # define a polygon collection based on individual cell polygons
    points = np.multiply(cells.cell_verts, p.um)
    collection =  PolyCollection(points, cmap=clrmap, edgecolors='none')
    collection.set_array(data)
    ax.add_collection(collection)

    return collection, ax



#!/usr/bin/env python3
# --------------------( LICENSE                           )--------------------
# Copyright 2014-2025 by Alexis Pietak & Cecil Curry.
# See "LICENSE" for further details.

'''
Unit tests exercising various simpler classes defined by the
:mod:`betse.util.type.iterable.mapping.mapcls` submodule *not* warranting their
own dedicated unit test submodule.
'''

# ....................{ IMPORTS                           }....................
from collections import OrderedDict
import pytest

# ....................{ TESTS ~ default                   }....................
def test_defaultdict_pass() -> None:
    '''
    Test aspects of the
    :class:`betse.util.type.iterable.mapping.mapcls.DefaultDict` type intended
    to succeed.
    '''

    # Defer heavyweight imports.
    from betse.util.type.iterable.mapping.mapcls import DefaultDict

    # Default dictionary to be tested.
    lake_isle_of_innisfree = DefaultDict(
        missing_key_value=lambda self, missing_key:
            missing_key + ' the deep heart’s core.',
        initial_mapping={
            'There midnight’s all': 'a glimmer',
            'and noon':             'a purple glow,',
            'And evening full':     'of the linnet’s wings.',
        },
    )

    # Test whether this dictionary contains a non-missing key initialized above.
    assert lake_isle_of_innisfree['and noon'] == 'a purple glow,'

    # Test whether a missing key defaults to the expected value.
    assert lake_isle_of_innisfree['I hear it in'] == (
        'I hear it in the deep heart’s core.')

# ....................{ TESTS ~ ordered                   }....................
def test_orderedargsdict_pass() -> None:
    '''
    Test aspects of the
    :class:`betse.util.type.iterable.mapping.mapcls.OrderedArgsDict` type
    intended to succeed.
    '''

    # Defer heavyweight imports.
    from betse.util.type.iterable.mapping.mapcls import OrderedArgsDict

    # Custom ordered dictionary to be tested.
    tuatha_de_danann = OrderedArgsDict(
        'Nuada', 'Nodens',
        'Lugh', 'Lugus',
        'Brigit', 'Brigantia',
    )

    # Standard ordered dictionary containing the same key-value pairs.
    tuath_de = OrderedDict((
        ('Nuada', 'Nodens'),
        ('Lugh', 'Lugus'),
        ('Brigit', 'Brigantia'),
    ))

    # Test whether these dictionaries contain the same key-value pairs.
    assert tuatha_de_danann == tuath_de

    # Test whether initializing an empty custom ordered dictionary succeeds.
    empty_dict = OrderedArgsDict()

    # Test whether this dictionary is indeed empty.
    assert not empty_dict


def test_orderedargsdict_fail() -> None:
    '''
    Test aspects of the
    :class:`betse.util.type.iterable.mapping.mapcls.OrderedArgsDict` type
    intended to fail.
    '''

    # Defer heavyweight imports.
    from betse.exceptions import BetseMappingException
    from betse.util.type.iterable.mapping.mapcls import OrderedArgsDict

    # Test whether initializing this dictionary with an odd rather than even
    # number of positional arguments fails.
    with pytest.raises(BetseMappingException):
        OrderedArgsDict(
            'Danu', 'Don',
            'Dian Cecht',
        )



#!/usr/bin/env python3
# Copyright 2014-2025 by Alexis Pietak & Cecil Curry.
# See "LICENSE" for further details.

'''
Layer subclasses spatially overlaying vector components as quiver plots onto the
cell cluster.
'''

# ....................{ IMPORTS                            }....................
from betse.science.math import mathunit
from betse.science.visual.layer.vectorfield.lyrvecfldabc import (
    LayerCellsFieldColorlessABC)
from betse.util.type.decorator.deccls import abstractproperty
# from betse.util.type.types import type_check
from numpy import ndarray

# ....................{ SUPERCLASSES                       }....................
class LayerCellsFieldQuiverABC(LayerCellsFieldColorlessABC):
    '''
    Layer abstract subclass plotting the most significant X and Y components of
    a single vector field spatially situated according to this layer subclass
    onto the cell cluster for one on more time steps.

    Attributes
    ----------
    _quiver_plot : matplotlib.quiver.Quiver
        Ouiver plot of all vector components previously plotted for the prior
        time step if any *or* ``None`` otherwise.
    '''

    # ..................{ INITIALIZERS                       }..................
    def __init__(self, *args, **kwargs) -> None:

        # Initialize our superclass.
        super().__init__(*args, **kwargs)

        # Default all remaining instance variables.
        self._quiver_plot = None

    # ..................{ SUPERCLASS                         }..................
    def _layer_next(self) -> None:

        # Replace all X and Y components of this vector field plotted for the
        # prior time step by those for this time step.
        self._quiver_plot.set_UVC(U=self._field_time_x, V=self._field_time_y)

    # ..................{ SUBCLASS                           }..................
    @abstractproperty
    def _field_time_x(self) -> ndarray:
        '''
        One-dimensional Numpy array of all X components of this vector field
        spatially situated spatially situated according to this layer subclass
        for the current time step.
        '''

        pass


    @abstractproperty
    def _field_time_y(self) -> ndarray:
        '''
        One-dimensional Numpy array of all Y components of this vector field
        spatially situated spatially situated according to this layer subclass
        for the current time step.
        '''

        pass

# ....................{ SUBCLASSES                         }....................
class LayerCellsFieldQuiverCells(LayerCellsFieldQuiverABC):
    '''
    Layer subclass plotting the most significant X and Y components of a single
    vector field spatially situated at cell centres (e.g., intracellular
    electric field) onto the cell cluster for one on more time steps.
    '''

    # ..................{ SUPERCLASS                         }..................
    def _layer_first(self) -> None:

        # Ouiver plot of all vector components plotted for this time step. See
        # the matplotlib.quiver.quiver() docstring for further details.
        self._quiver_plot = self._visual.axes.quiver(
            # Positional arguments. Thanks to internal flaws in the
            # matplotlib.quiver._parse_args() function parsing arguments passed
            # to the matplotlib.axes.quiver() method called here, the first four
            # arguments *MUST* be passed as positional arguments.

            # Upscaled X and Y coordinates of all cell centres.
            self._phase.cache.upscaled.cells_centre_x,
            self._phase.cache.upscaled.cells_centre_y,

            # Normalized X and Y components of this vector field spatially
            # situated at cell centres for this time step.
            self._field_time_x,
            self._field_time_y,

            # Keyword arguments. All remaining arguments *MUST* be passed as
            # keyword arguments.

            # Matplotlib-specific color code of all vector arrows.
            color=self._phase.p.vcolor,

            # Multiples of the width and height (respectively) of vector arrow
            # shafts by which to scale the width and height of vector arrow
            # heads. These settings default to 3 and 5 (respectively).
            headwidth=5,
            headlength=7,

            # The portion of each vector arrow to situate at the X and Y
            # coordinates of the corresponding cell centre.
            pivot='middle',

            # Scale vector arrows such that arrow size increases as the
            # user-defined zoom level increases in either X or Y dimensions.
            # units='xy',
            units='x',

            # Z-order of this plot with respect to other artists.
            zorder=self._zorder,
        )

    # ..................{ SUPERCLASS ~ property              }..................
    @property
    def _field_time_x(self) -> ndarray:
        return self._field.times_cells_centre.unit_x[self._visual.time_step]

    @property
    def _field_time_y(self) -> ndarray:
        return self._field.times_cells_centre.unit_y[self._visual.time_step]


class LayerCellsFieldQuiverGrids(LayerCellsFieldQuiverABC):
    '''
    Layer subclass plotting the most significant X and Y components of a single
    vector field spatially situated at environmental grid space centres (e.g.,
    extracellular electric field) onto the cell cluster for one on more time
    steps.
    '''

    # ..................{ SUPERCLASS                         }..................
    def _layer_first(self) -> None:

        # Ouiver plot of all vector components plotted for this time step. See
        # the matplotlib.quiver.quiver() docstring for further details.
        self._quiver_plot = self._visual.axes.quiver(
            # Positional arguments. See
            # LayerCellsFieldQuiverCells._layer_first() for further discussion.

            # Upscaled X and Y coordinates of all grid space centres.
            self._phase.cache.upscaled.grids_centre_x,
            self._phase.cache.upscaled.grids_centre_y,

            # Normalized X and Y components of this vector field spatially
            # situated at grid space centres for this time step.
            self._field_time_x,
            self._field_time_y,

            #FIXME: DRY. All following keyword arguments are duplicated from the
            #LayerCellsFieldQuiverCells._layer_first() method, which is bad. In
            #fact, the only difference between this and that class is the use
            #of "grids_centre"-centric arrays above and below. So, four
            #differences in total. Can we generalize this further?

            # Keyword arguments. All remaining arguments *MUST* be passed as
            # keyword arguments.

            # Matplotlib-specific color code of all vector arrows.
            color=self._phase.p.vcolor,

            # Multiples of the width and height (respectively) of vector arrow
            # shafts by which to scale the width and height of vector arrow
            # heads. These settings default to 3 and 5 (respectively).
            headwidth=5,
            headlength=7,

            # The portion of each vector arrow to situate at the X and Y
            # coordinates of the corresponding cell centre.
            pivot='middle',

            # Scale vector arrows such that arrow size increases as the
            # user-defined zoom level increases in either X or Y dimensions.
            # units='xy',
            units='x',

            # Z-order of this plot with respect to other artists.
            zorder=self._zorder,
        )

    # ..................{ SUPERCLASS ~ property              }..................
    @property
    def _field_time_x(self) -> ndarray:

        field = self._field.times_grids_centre
        return (
            field.x[self._visual.time_step] /
            field.magnitudes[self._visual.time_step].max())


    @property
    def _field_time_y(self) -> ndarray:

        field = self._field.times_grids_centre
        return (
            field.y[self._visual.time_step] /
            field.magnitudes[self._visual.time_step].max())


class LayerCellsFieldQuiverMembranes(LayerCellsFieldQuiverABC):
    '''
    Layer subclass plotting the most significant X and Y components of a single
    vector field spatially situated at cell membrane midpoints (e.g.,
    microtubules) onto the cell cluster for one on more time steps.

    Attributes
    ----------
    _cells_radius : ndarray
        One-dimensional Numpy array indexing each cell such that each element is
        the upscaled radius of that cell.
    '''

    # ..................{ INITIALIZERS                       }..................
    def __init__(self, *args, **kwargs) -> None:

        # Initialize our superclass.
        super().__init__(*args, **kwargs)

        # Default all remaining instance variables.
        self._cells_radius = None


    def prep(self, *args, **kwargs) -> None:

        # Prepare our superclass.
        super().prep(*args, **kwargs)

        # Upscale all cell radii.
        self._cells_radius = mathunit.upscale_coordinates(
            self._phase.cells.R[self._phase.cells.mem_to_cells])

    # ..................{ SUPERCLASS                         }..................
    def _layer_first(self) -> None:

        # Localize attributes for brevity.
        cells = self._phase.cells

        # Ouiver plot of all vector components plotted for this time step. See
        # the matplotlib.quiver.quiver() docstring for further details.
        self._quiver_plot = self._visual.axes.quiver(
            # Positional arguments. See
            # LayerCellsFieldQuiverCells._layer_first() for further discussion.

            # Upscaled X and Y coordinates of all cell membrane midpoints.
            mathunit.upscale_coordinates(
                cells.cell_centres[:, 0][cells.mem_to_cells]),
            mathunit.upscale_coordinates(
                cells.cell_centres[:, 1][cells.mem_to_cells]),

            # X and Y components of this vector field spatially situated at cell
            # membrane midpoints extending to cell centres for this time step.
            self._field_time_x,
            self._field_time_y,

            # Keyword arguments. All remaining arguments *MUST* be passed as
            # keyword arguments.

            # Matplotlib-specific color code of all vector arrows.
            color=self._phase.p.vcolor,

            # Number of data units per arrow length unit.
            scale=mathunit.upscale_coordinates(self._phase.p.wsx * 0.8),

            # Z-order of this plot with respect to other artists.
            zorder=self._zorder,
        )

    # ..................{ SUPERCLASS ~ property              }..................
    @property
    def _field_time_x(self) -> ndarray:
        '''
        X components of this vector field spatially situated at cell membrane
        midpoints extending exactly to cell centres for the current time step,
        calculated by first normalizing and then extending each such component
        by the upscaled radius of the cell containing this membrane.
        '''

        return self._field.times_membranes_midpoint.unit_x[
            self._visual.time_step] * self._cells_radius


    @property
    def _field_time_y(self) -> ndarray:
        '''
        Y components of this vector field spatially situated at cell membrane
        midpoints extending exactly to cell centres for the current time step.
        '''

        return self._field.times_membranes_midpoint.unit_y[
            self._visual.time_step] * self._cells_radius



#!/usr/bin/env python3
# --------------------( LICENSE                            )--------------------
# Copyright 2014-2025 by Alexis Pietak & Cecil Curry.
# See "LICENSE" for further details.

'''
Main entry point of this application's command line interface (CLI).

This submodule is a thin wrapper intended to be:

* Indirectly imported and run from external entry point scripts installed by
  setuptools (e.g., the ``betse`` command).
* Directly imported and run from the command line (e.g., via
  ``python -m betse.cli``).
'''

# ....................{ IMPORTS                            }....................
from beartype.typing import Collection
from betse.util.os.command import cmdexit

# ....................{ MAIN                               }....................
def main(arg_list: Collection[str] = None) -> int:
    '''
    Run this application's command-line interface (CLI) with the passed
    arguments if non-``None`` *or* with the arguments passed on the command
    line (i.e., :attr:`sys.argv`) otherwise.

    This function is provided as a convenience to callers requiring procedural
    functions rather than conventional methods (e.g., :mod:`setuptools`).

    Parameters
    ----------
    arg_list : Collection[str]
        Collection of zero or more string arguments to pass to this interface.
        Defaults to :data:`None`, in which case arguments passed on the command
        line (i.e., :attr:`sys.argv`) will be used instead.

    Returns
    -------
    int
        Exit status of this interface and hence this process as an unsigned
        byte (i.e., integer in the range ``[0, 255]``).
    '''

    # Defer function-specific imports.
    from betse.appmeta import BetseAppMeta
    from betse.cli.climain import BetseCLI

    # Initialize this application (excluding mandatory dependencies of this
    # application, which will be initialized later in the startup process) by
    # instantiating the BETSE-specific application metadata singleton.
    BetseAppMeta()

    # Implement this application by instantiating and running the
    # BETSE-specific command-line interface (CLI).
    return BetseCLI().run(arg_list)

# ....................{ MAIN                              }....................
# If this module is imported from the command line, run this application's CLI;
# else, noop. For POSIX compliance, the exit status returned by this function
# is propagated to the caller as this script's exit status.
if __name__ == '__main__':
    cmdexit.exit_with_status(main())



#!/usr/bin/env python3
# --------------------( LICENSE                           )--------------------
# Copyright 2014-2025 by Alexis Pietak & Cecil Curry.
# See "LICENSE" for further details.

'''
Low-level logging formatter subclasses.
'''

# ....................{ IMPORTS                           }....................
#!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
# WARNING: To avoid circular import dependencies, avoid importing from *ANY*
# application-specific modules at the top-level -- excluding those explicitly
# known *NOT* to import from this module. Since all application-specific
# modules must *ALWAYS* be able to safely import from this module at any level,
# these circularities are best avoided here rather than elsewhere.
#!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!

# from betse.util.type.types import type_check
from logging import Formatter

# ....................{ CLASSES                           }....................
#FIXME: Unfortunately, this fundamentally fails to work. The reason why? The
#"TextWrapper" class inserts spurious newlines *EVEN WHEN YOU EXPLICITLY TELL
#IT NOT TO*. This is crazy, but noted in the documentation:
#
#    "If replace_whitespace is False, newlines may appear in the middle of a
#     line and cause strange output. For this reason, text should be split into
#     paragraphs (using str.splitlines() or similar) which are wrapped
#     separately."
#
#Until this is resolved, the only remaining means of wrapping log messages will
#be to define new top-level module functions suffixed by "_wrapped" ensuring
#that the appropriate formatter is used (e.g., a new log_info_wrapped()
#function). For now, let's just avoid the topic entirely. It's all a bit
#cumbersome and we're rather weary of it.

class LogFormatterWrap(Formatter):
    '''
    Log formatter wrapping all lines in handled log records to a sane line
    length (e.g., 80 characters).

    Attributes
    ----------
    _text_wrapper : TextWrapper
        Object with which to wrap log messages, cached for efficiency.
    '''

    pass
    # def __init__(self, *args, **kwargs):
    #     super().__init__(*args, **kwargs)
    #     self._text_wrapper = TextWrapper(
    #         drop_whitespace = False,
    #         replace_whitespace = False,
    #     )

    # def format(self, log_record: LogRecord) -> str:
    #     # Avoid circular import dependencies.
    #     from betse.util.type import strs
    #
    #     # Get such message by (in order):
    #     #
    #     # * Formatting such message according to our superclass.
    #     # * Wrapping such formatted message.
    #     return strs.wrap(
    #         text = super().format(log_record),
    #         text_wrapper = self._text_wrapper,
    #     )



#!/usr/bin/env python3
# --------------------( LICENSE                           )--------------------
# Copyright 2014-2025 by Alexis Pietak & Cecil Curry.
# See "LICENSE" for further details.

'''
Low-level **iterator** (i.e., objects satisfying the standard iterator API)
facilities.
'''

# ....................{ IMPORTS                           }....................
from betse.util.type.types import IterableTypes

# ....................{ GENERATORS                        }....................
def empty_iterator() -> IterableTypes:
    '''
    Empty iterator iterating over... absolutely nothing.

    See Also
    ----------
    https://stackoverflow.com/a/10621647/2809027
        StackOverflow answer strongly inspiring this implementation.
    '''

    # It is clever. It is Python.
    return iter(())



#!/usr/bin/env python3
# --------------------( LICENSE                            )--------------------
# Copyright 2014-2025 by Alexis Pietak & Cecil Curry.
# See "LICENSE" for further details.

'''
High-level vector field subcache functionality.
'''

# ....................{ IMPORTS                            }....................
from betse.lib.numpy import nparray
from betse.science.math import mathunit
from betse.science.math.cache.cacheabc import SimPhaseCacheABC
from betse.science.math.vector.veccls import VectorCellsCache
from betse.science.math.vector.vecfldcls import VectorFieldCellsCache
from betse.util.type.decorator.decmemo import property_cached

# FIXME: Sess, why the heck would you put these plotting functions into math?!?!?

# ....................{ SUBCLASSES                         }....................
class SimPhaseCacheVectorFieldCells(SimPhaseCacheABC):
    '''
    Simulation phase-specific vector field subcache, persisting all previously
    constructed vector fields for a single simulation phase.
    '''

    # ..................{ PROPERTIES ~ currents              }..................
    @property_cached
    def currents_intra(self) -> VectorFieldCellsCache:
        '''
        Vector field cache of all intracellular current densities over all time
        steps of the current simulation phase, originally spatially situated at
        cell centres.

        For readability of units in exported visuals (e.g., plots), this cache
        additionally upscales these densities to units of uA/cm^2.
        '''

        return VectorFieldCellsCache(
            x=VectorCellsCache(
                phase=self._phase,
                times_cells_centre=mathunit.upscale_units_centi(
                    self._phase.sim.I_cell_x_time)),
            y=VectorCellsCache(
                phase=self._phase,
                times_cells_centre=mathunit.upscale_units_centi(
                    self._phase.sim.I_cell_y_time)),
        )


    @property_cached
    def currents_extra(self) -> VectorFieldCellsCache:
        '''
        Vector field cache of all exracellular current densities over all time
        steps of the current simulation phase, originally spatially situated at
        environmental grid space centres.

        For readability of units in exported visuals (e.g., plots), this cache
        additionally upscales these densities to units of uA/cm^2.
        '''

        # Create, return, and cache this field.
        return VectorFieldCellsCache(
            x=VectorCellsCache(
                phase=self._phase,
                times_grids_centre=mathunit.upscale_units_centi(
                    self._phase.sim.I_tot_x_time)),
            y=VectorCellsCache(
                phase=self._phase,
                times_grids_centre=mathunit.upscale_units_centi(
                    self._phase.sim.I_tot_y_time)),
        )

    # ..................{ PROPERTIES ~ deform                }..................
    #FIXME: Raise an exception unless deformations are enabled. To do so sanely,
    #we'll want to define a new @phase_property_cached decorator accepting an
    #optional "requirements" parameter, much like the existing @piperunner
    #decorator. (For now, simply ignore this for simplicity.)
    @property_cached
    def deform_total(self) -> VectorFieldCellsCache:
        '''
        Vector field cache of all upscaled **total cellular displacements**
        (i.e., summations of all cellular deformations due to galvanotropic and
        osmotic pressure body forces) over all sampled time steps of the current
        simulation phase, originally spatially situated at cell centres.

        For readability of units in exported visuals (e.g., plots), this cache
        additionally upscales these displacements from meters to micrometers.
        '''

        return VectorFieldCellsCache(
            x=VectorCellsCache(
                phase=self._phase,
                times_cells_centre=mathunit.upscale_coordinates(
                    self._phase.sim.dx_cell_time)),
            y=VectorCellsCache(
                phase=self._phase,
                times_cells_centre=mathunit.upscale_coordinates(
                    self._phase.sim.dy_cell_time)),
        )

    # ..................{ PROPERTIES ~ electric              }..................
    @property_cached
    def electric_intra(self) -> VectorFieldCellsCache:
        '''
        Vector field cache of the intracellular electric field over all sampled
        time steps of the current simulation phase, originally spatially
        situated at cell membrane midpoints.
        '''

        return VectorFieldCellsCache(
            x=VectorCellsCache(
                phase=self._phase,
                times_membranes_midpoint=self._phase.sim.efield_gj_x_time),
            y=VectorCellsCache(
                phase=self._phase,
                times_membranes_midpoint=self._phase.sim.efield_gj_y_time),
        )


    @property_cached
    def electric_extra(self) -> VectorFieldCellsCache:
        '''
        Vector field cache of the extracellular electric field over all sampled
        time steps of the current simulation phase, originally spatially
        situated at environmental grid space centres.

        Raises
        ----------
        BetseSimConfException
            If this simulation has disabled extracellular spaces.
        '''

        # If extracellular spaces are disabled, raise an exception.
        self._phase.p.die_unless_ecm()

        # Create, return, and cache this field.
        return VectorFieldCellsCache(
            x=VectorCellsCache(
                phase=self._phase,
                times_grids_centre=self._phase.sim.efield_ecm_x_time),
            y=VectorCellsCache(
                phase=self._phase,
                times_grids_centre=self._phase.sim.efield_ecm_y_time),
        )

    # # ..................{ PROPERTIES ~ microtubule           }..................
    # @property_cached
    # def microtubule(self) -> VectorFieldCellsCache:
    #     '''
    #     Vector field cache of all cellular microtubules over all time steps of
    #     the current simulation phase, originally spatially situated at cell
    #     membrane midpoints.
    #     '''
    #
    #     return VectorFieldCellsCache(
    #         x=VectorCellsCache(
    #             phase=self._phase,
    #             times_membranes_midpoint=self._phase.sim.mtubes_x_time),
    #         y=VectorCellsCache(
    #             phase=self._phase,
    #             times_membranes_midpoint=self._phase.sim.mtubes_y_time),
    #     )

    # ..................{ PROPERTIES ~ voltage               }..................
    @property_cached
    def voltage_polarity(self) -> VectorFieldCellsCache:
        '''
        Vector field cache of all cellular voltage polarities over all time
        steps of the current simulation phase, originally spatially situated at
        cell membrane midpoints.
        '''

        # Two-dimensional Numpy arrays of all transmembrane voltages (Vmem) and
        # Vmem averages across all cell membranes over all time steps.
        vm_time     = nparray.from_iterable(self._phase.sim.vm_time)
        vm_ave_time = nparray.from_iterable(self._phase.sim.vm_ave_time)

        # Two-dimensional Numpy array of all transmembrane voltage polarity
        # vector magnitudes whose:
        #
        # * First dimension indexes each time step.
        # * Second dimension indexes each cell membrane such that each element
        #   is the magnitude of the polarity of the transmembrane voltage across
        #   that membrane for this time step, where this magnitude is defined as
        #   the difference between:
        #   * The transmembrane voltage across that membrane for this time step.
        #   * The average transmembrane voltage across all membranes of the cell
        #     containing that membrane for this time step.
        polarity_membranes_midpoint_magnitudes = (
            vm_time - vm_ave_time[:,self._phase.cells.mem_to_cells])

        # Two-dimensional Numpy arrays of the X and Y components of all Vmem
        # polarity vectors, spatially situated at cell membrane midpoints.
        polarity_membranes_midpoint_x = (
            polarity_membranes_midpoint_magnitudes *
            self._phase.cells.membranes_normal_unit_x)
        polarity_membranes_midpoint_y = (
            polarity_membranes_midpoint_magnitudes *
            self._phase.cells.membranes_normal_unit_y)

        # Two-dimensional Numpy arrays of the X and Y components of all Vmem
        # polarity vectors, spatially situated at cell centres.
        polarity_cells_centre_x = (
            self._phase.cells.map_membranes_midpoint_to_cells_centre(
                polarity_membranes_midpoint_x * self._phase.cells.mem_sa) /
            self._phase.cells.cell_sa)
        polarity_cells_centre_y = (
            self._phase.cells.map_membranes_midpoint_to_cells_centre(
                polarity_membranes_midpoint_y * self._phase.cells.mem_sa) /
            self._phase.cells.cell_sa)

        # Create, return, and cache this vector field.
        return VectorFieldCellsCache(
            x=VectorCellsCache(
                phase=self._phase, times_cells_centre=polarity_cells_centre_x),
            y=VectorCellsCache(
                phase=self._phase, times_cells_centre=polarity_cells_centre_y))



#!/usr/bin/env python3
# --------------------( LICENSE                            )--------------------
# Copyright 2014-2025 by Alexis Pietak & Cecil Curry.
# See "LICENSE" for further details.

'''
Primitive two-dimensional point functionality.
'''

# ....................{ IMPORTS                            }....................
from betse.exceptions import BetseMathLineException, BetseMathPointException
from betse.util.type.iterable import iterables
from betse.util.type.types import (
    type_check, BoolTypes, NumericTypes, NumpyArrayType, SequenceTypes)

# ....................{ EXCEPTIONS                         }....................
@type_check
def die_unless_point(*points: SequenceTypes) -> None:
    '''
    Raise an exception unless all passed sequences are two-dimensional points.

    Parameters
    ----------
    points : tuple[SequenceTypes]
        Tuple of all sequences to be validated.

    Raises
    ----------
    BetseMathPointException
        If any such sequence is *not* a two-dimensional point.

    See Also
    ----------
    :func:`is_point`
        Further details.
    '''

    # If one or more of the passed objects are *NOT* points...
    if not is_point(*points):
        # For each such object...
        for point in points:
            # If this object is *NOT* a point...
            if not is_point(point):
                # If this object does *NOT* contain a coordinate pair, raise an
                # appropriate exception.
                if len(point) != 2:
                    raise BetseMathPointException(
                        'Sequence not a two-dimensional point '
                        '(i.e., length != 2): {!r}'.format(point))
                # Else, this object contains a coordinate pair. By definition
                # of the is_point() function, it must be the case that one or
                # more of these coordinates are *NOT* numbers. In this case,
                # raise an appropriate exception.
                else:
                    raise BetseMathPointException(
                        'Sequence not a two-dimensional point '
                        '(i.e., one or more coordinates '
                        'not numbers): {!r}'.format(point))

# ....................{ TESTERS                            }....................
@type_check
def is_point(*points: SequenceTypes) -> bool:
    '''
    ``True`` only if all passed sequences are **two-dimensional points** (i.e.,
    contain exactly two numbers providing the X and Y coordinates of these
    points).

    Parameters
    ----------
    points : tuple[SequenceTypes]
        Tuple of all sequences to be tested.

    Returns
    ----------
    bool
        ``True`` only if these sequences are all two-dimensional points.
    '''

    # Return true only if...
    return all(
        # This point contains exactly two coordinates *AND*...
        len(point) == 2 and (
            # This point is either a Numpy array presumably containing numbers
            # *OR* a Python sequence such that...
            isinstance(point, NumpyArrayType) or all(
                # This coordinate is a number...
                isinstance(point_coordinate, NumericTypes)
                # For each coordinate of this point.
                for point_coordinate in point
            )
        )
        # For each passed point.
        for point in points)


@type_check
def is_left_of_vector(
    subject_point: SequenceTypes,
    vector_head_point: SequenceTypes,
    vector_tail_point: SequenceTypes,
) -> BoolTypes:
    '''
    ``True`` only if the passed two-dimensional subject point is spatially
    situated to the left of the two-dimensional vector defined by the passed
    two-dimensional head and tail points.

    Equivalently, this function returns ``True`` only if:

    * This vector is spatially situated to the right of this subject point.
    * A clockwise rotation rotates this subject point onto this vector.
    * A counter-clockwise rotation rotates this vector onto this subject point.
    * The determinant and hence sign of the cross product of this vector and
      this **subject vector** (i.e., vector whose head is this subject point
      and tail is the passed tail point) is positive.

    Derivation
    ----------
    We now derive the implementation of this function from the prior statement.
    If the angle between this vector and this subject vector (in that order)
    is:

    * Positive, a counter-clockwise rotation rotates the former onto the
      latter. In this case, this vector is spatially situated to the right of
      this subject point.
    * Negative, a clockwise rotation rotates the former onto the latter. In
      this case, this vector is spatially situated to the left of this subject
      point.

    This function thus reduces to testing the sign of the angle between this
    vector and this subject vector. While there exist numerous means of doing
    so, the most computationally efficient in both space and time is to test
    the sign of the determinant of the cross product of these vectors --
    requiring exactly four subtractions, two multiplications, and one
    comparison.

    The traditional definition of the cross product ``u × v`` of two arbitrary
    vectors ``u`` and ``v`` is given by the following two equivalent
    equalities:

    .. code::

       u × v = |u||v|sin(θ) n
               | i  j  k|
               |ux uy uz|
       u × v = |vx vy vz|

    Where:

    * ``|u|`` and ``|v|`` are the magnitudes of these vectors.
    * ``θ`` is the angle between these vectors.
    * ``n`` is a unit vector perpendicular to the plane containing these
      vectors in the direction oriented by the right-hand rule.
    * ``i``, ``j``, and ``k`` are the standard basis vectors.
    * ``ux``, ``uy``, and ``uz`` are the X, Y, and Z components of ``u``: e.g.,
      ``u = ux i + uy j + uz k``.
    * ``vx``, ``vy``, and ``vz`` are the X, Y, and Z components of ``v``: e.g.,
      ``v = vx i + vy j + vz k``.

    In this case, both ``u`` and ``v`` are two-dimensional, so:

    * ``n = k`` (i.e., ``u × v`` is a vector purely in the Z dimension).
    * ``uz = vz = 0``.

    By cofactor expansion, the latter equality for ``u × v`` then expands to:

    .. code::

               |uy uz|    |ux uz|    |ux uy|
       u × v = |vy vz|i - |vx vz|j + |vx vy|k
       u × v = (uy·vz - uz·vy)i - (ux·vz - uz·vx)j + (ux·vy - uy·vx)k
       u × v = (ux·vy - uy·vx)k

    By definition, the determinant ``det(u × v)`` of this cross product may be
    shown to similarly reduce to:

    .. code::

       det(u × v) = ux·vy - uy·vx

    The latter equality for ``u × v`` then reduces to:

    .. code::

       u × v = det(u × v) k

    The sign of the angle ``θ`` between ``u`` and ``v`` is then obtained by
    equating these two equalities and solving for ``θ`` as follows:

    .. code::

       |u||v|sin(θ) k = det(u × v) k
                det(u × v)
                ----------
       sin(θ) = |u||v|
                 [det(u × v)]
                 [----------]
       θ = arcsin[|u||v|    ]

    This computationally expensive equality may be reduced by noting that we
    only require the sign of ``θ``; the actual value of ``θ`` is irrelevant. In
    particular:

    * The ``arcsin()`` function is a monotonically increasing function vaguely
      resembling a wavy varient of the simple line ``y = x``, implying the sign
      of this function is *always* the sign of its operand, implying this
      function is ignorable with respect to this sign.
    * The quantity ``1/(|u||v|)`` is *always* positive, implying this quantity
      is also ignorable with respect to this sign.

    The sign of ``θ`` is thus the sign of ``det(u × v)``, the only remaining
    quantity with unknown sign in the above equality. This sign is positive if
    the following conditionality holds:

    .. code::

       ux·vy - uy·vx = det(u × v) > 0
       ux·vy > uy·vx

    Since this function accepts points rather than vectors, the former must be
    translated to the latter. Let:

    * ``h`` be the passed head point.
    * ``t`` be the passed tail point.
    * ``p`` be the passed subject point.
    * ``v`` be the corresponding subject vector.

    The X and Y components of ``u`` and ``v``  may then be defined in terms of
    these points as follows:

    * ``ux = hx - tx``.
    * ``uy = hy - ty``.
    * ``vx = px - tx``.
    * ``vy = py - ty``.

    The above conditionality then expands in terms of these points to:

    .. code::

       (hx - tx)·(py - ty) > (hy - ty)·(px - tx)

    The implementation of this function trivially follows.

    Parameters
    ----------
    vector_head_point : SequenceTypes
        2-sequence of the X and Y coordinates of the two-dimensional head point
        of this vector such that:

        * The first item is the X coordinate of this point.
        * The second item is the Y coordinate of this point.
    vector_tail_point : SequenceTypes
        2-sequence of the X and Y coordinates of the two-dimensional tail point
        of this vector.
    subject_point : SequenceTypes
        2-sequence of the X and Y coordinates of the two-dimensional **subject
        point** (i.e., point to test the orientation of this vector against).

    Returns
    ----------
    BoolTypes
        True only if this vector is to the right of this subject point. Note
        that this is a Numpy-style boolean if the passed sequences are Numpy
        arrays rather than builtin Python objects.
    '''

    # If any passed sequence is *NOT* a point, raise an exception.
    die_unless_point(subject_point, vector_head_point, vector_tail_point)
    # die_unless_point(subject_point)
    # die_unless_point(vector_head_point)
    # die_unless_point(vector_tail_point)

    # See the above derivation for details. Math, you win all the efficiency.
    return (
        (vector_head_point[0] - vector_tail_point[0])*
            (subject_point[1] - vector_tail_point[1]) >
        (vector_head_point[1] - vector_tail_point[1])*
            (subject_point[0] - vector_tail_point[0])
    )

# ....................{ INTERSECTERS                       }....................
@type_check
def intersect_lines(
    line1_point1: SequenceTypes,
    line1_point2: SequenceTypes,
    line2_point1: SequenceTypes,
    line2_point2: SequenceTypes,
) -> SequenceTypes:
    '''
    Two-dimensional point intersecting the pair of lines defined by the passed
    pairs of two-dimensional points if any *or* raise an exception if these
    lines either infinitely overlap (i.e., are collinear) or never overlap
    (i.e., are parallel).

    Derivation
    ----------
    We now derive the implementation of this function from the two-point form
    of a two-dimensional line. Let:

    * ``c`` be the first passed point defining the first line.
    * ``d`` be the second passed point defining the first line.
    * ``s`` be the first passed point defining the second line.
    * ``t`` be the second passed point defining the second line.
    * ``e`` be any point residing on the first line.
    * ``u`` be any point residing on the second line.

    The intersection of these lines if any is the point satisfying the equality
    ``e = u``. For simplicity, let ``x = ex = ux`` and ``y = ex = ux`` be the
    X and Y coordinates of this point. The two-point form of the lines defined
    by these points constrained to intersect at this point is as follows:

    .. code::

                (dy - cy)(x - cx)
                -----------------
       y - cy =  dx - cx
                (ty - sy)(x - sx)
                ------------------
       y - sy =  tx - sx

    Determining this intersection point if any reduces to solving this system
    of two linear equations of two variables, which typically proceeds as
    follows:

    .. code::

       (y - cy)(dx - cx) = (dy - cy)(x - cx)
       (y - sy)(tx - sx) = (ty - sy)(x - sx)

       y·dx - y·cx - cy·dx + cy·cx = x·dy - dy·cx - x·cy + cy·cx
       y·tx - y·sx - sy·tx + sy·sx = x·ty - ty·sx - x·sy + sy·sx

       (cy - dy)x + (dx - cx)y = dx·cy - dy·cx
       (sy - ty)x + (tx - sx)y = tx·sy - ty·sx

    Let ``M`` be the coefficient matrix describing this linear system of
    equations such that:

    .. code::

           |cy-dy  dx-cx|
       M = |sy-ty  tx-sx|

    The determinant ``det(M)`` of this matrix is given by:

    .. code::

       det(M) = (cy-dy)·(tx-sx) - (dx-cx)·(sy-ty)

    This linear system of equations is then equivalent to this matrix equation:

    .. code::

        |x|   |dx·cy-dy·cx|
       M|y| = |tx·sy-ty·sx|

    This linear system of equations has a unique solution (implying these lines
    to intersect in a single point) if and only if:

    .. code::

       det(M) = (cy-dy)·(tx-sx) -  (dx-cx)·(sy-ty) != 0
                (cy-dy)·(tx-sx) != (dx-cx)·(sy-ty)

    Suppose ``det(M) != 0``, in which case this linear system of equastions has
    a unique solution. While there exist various (equally valid) means of
    obtaining this solution, the simplest is by invocation of Cramer's Rule.
    Let ``Mx`` be the matrix formed by replacing the first column of the
    coefficient matrix with the constant column vector and ``My`` the matrix
    formed by replacing the second column of the coefficient matrix with the
    constant column vector such that:

    .. code::

            |dx·cy-dy·cx  dx-cx      |
       Mx = |tx·sy-ty·sx  tx-sx      |
            |cy-dy        dx·cy-dy·cx|
       My = |sy-ty        tx·sy-ty·sx|

    The determinants ``det(Mx)`` and ``det(My)`` of these matrices are given
    by:

    .. code::

       det(Mx) = (dx·cy-dy·cx)(tx-sx) - (tx·sy-ty·sx)(dx-cx)
       det(My) = (tx·sy-ty·sx)(cy-dy) - (dx·cy-dy·cx)(sy-ty)

    Cramer's Rule then yields the X and Y coordinates of this point of
    intersection (with respect to these determinants) as follows:

    .. code::

           det(Mx)   (dx·cy-dy·cx)(tx-sx) - (tx·sy-ty·sx)(dx-cx)
           ------- = -------------------------------------------
       x = det(M)    (cy-dy)·(tx-sx) - (dx-cx)·(sy-ty)
           det(My)   (tx·sy-ty·sx)(cy-dy) - (dx·cy-dy·cx)(sy-ty)
           ------- = -------------------------------------------
       y = det(M)    (cy-dy)·(tx-sx) - (dx-cx)·(sy-ty)

    For efficiency, this solution is often rewritten into the equivalent form:

           (dx·cy-dy·cx)(tx-sx) - (tx·sy-ty·sx)(dx-cx)
           -------------------------------------------
       x = (dx-cx)·(ty-sy) - (dy-cy)·(tx-sx)
           (dx·cy-dy·cx)(ty-sy) - (tx·sy-ty·sx)(dy-cy)
           -------------------------------------------
       y = (dx-cx)·(ty-sy) - (dy-cy)·(tx-sx)

    The implementation of this function trivially follows.

    Parameters
    ----------
    line1_point1 : SequenceTypes
        2-sequence of the X and Y coordinates of the first two-dimensional
        point residing on the first of these lines (in any order) such that:

        * The first item is the X coordinate of this point.
        * The second item is the Y coordinate of this point.
    line1_point2 : SequenceTypes
        2-sequence of the X and Y coordinates of the second two-dimensional
        point residing on the first of these lines (in any order).
    line2_point1 : SequenceTypes
        2-sequence of the X and Y coordinates of the first two-dimensional
        point residing on the second of these lines (in any order).
    line2_point2 : SequenceTypes
        2-sequence of the X and Y coordinates of the second two-dimensional
        point residing on the second of these lines (in any order).

    Returns
    ----------
    SequenceTypes
        2-sequence of the X and Y coordinates of the two-dimensional point
        intersecting the pair of lines defined by these two-dimensional points.
        The type of this sequence is the same as that of the first passed
        point.

    Raises
    ----------
    BetseMathLineException
        If these lines either:

        * Infinitely overlap (i.e., are collinear).
        * Never overlap (i.e., are parallel).
    '''

    # If any passed sequence is *NOT* a point, raise an exception.
    die_unless_point(line1_point1, line1_point2, line2_point1, line2_point2)

    # X coordinates of these points in the nomenclature documented above.
    cx = line1_point1[0]
    dx = line1_point2[0]
    sx = line2_point1[0]
    tx = line2_point2[0]

    # Y coordinates of these points in the nomenclature documented above.
    cy = line1_point1[1]
    dy = line1_point2[1]
    sy = line2_point1[1]
    ty = line2_point2[1]

    # Run and rise of the slope of the first line.
    dcx = dx - cx
    dcy = dy - cy

    # Run and rise of the slope of the second line.
    tsx = tx - sx
    tsy = ty - sy

    # Determinant of the coefficient matrix solving this intersection.
    detM = dcx*tsy - dcy*tsx

    # If this determinant is zero, no unique solution exists, in which case
    # these lines either infinitely or never intersect. Since this constitutes
    # an error condition in either case, raise an exception. While we could
    # detect and differentiate these two error conditions in this exception
    # message, a lazy approach currently bests an accurate approach.
    if detM == 0:
        raise BetseMathLineException(
            'Line through points {!r} and {!r} and '
            'line through points {!r} and {!r} '
            'either infinitely or never intersect.'.format(
                cx, dx, sx, tx))
    # Else, this determinant is non-zero, in which a unique solution exists and
    # these lines uniquely intersect at a single point, which we now obtain.

    #                                  |dx dy|
    # Determinant of the ad-hoc matrix |cx cy|.
    detdc = dx*cy - dy*cx

    #                                  |tx ty|
    # Determinant of the ad-hoc matrix |sx sy|.
    detts = tx*sy - ty*sx

    # 2-tuple of the X and Y coordinates of this intersection, calculated via
    # Cramer's Rule from the requisite determinants.
    intersection_point = (
        (detdc*tsx - detts*dcx) / detM,
        (detdc*tsy - detts*dcy) / detM,
    )

    # This intersection as a sequence of the same type as the first passed
    # point.
    intersection_point_sequence = iterables.to_iterable(
        iterable=intersection_point, cls=type(line1_point1))

    # If this intersection is *NOT* a point, raise an exception.
    die_unless_point(intersection_point_sequence)

    # Return this point.
    return intersection_point_sequence



#!/usr/bin/env python3
# --------------------( LICENSE                           )--------------------
# Copyright 2014-2025 by Alexis Pietak & Cecil Curry.
# See "LICENSE" for further details.

'''
High-level :mod:`pymod`-based general-purpose decorators marking tests.

This submodule aggregates general-purpose decorators *not* conditionally
marking their decorated tests as either failed, parametrized, or skipped.
'''

# ....................{ IMPORTS                           }....................
import pytest

# ....................{ MARKS                             }....................
noop = pytest.mark.noop
'''
Preserve the decorated test or test parameter as is with *no* modification.

Caveats
----------
This decorator has the unintended side effect of also marking all tests and
test parameters decorated by this decorated with the ``noop`` tag. This
harmless albeit unfortunate side effect is the result of the :mod:`pytest.mark`
API, which strictly requires that *all* decorators passed to the ``marks``
parameter of the :mod:`pytest.param` function be strictly generated by the
:mod:`pytest.mark` API, which then imposes this seemingly arbitrary constraint.

In other words, there's absolutely nothing to see here, folks.
'''



#!/usr/bin/env python3
# Copyright 2014-2025 by Alexis Pietak & Cecil Curry.
# See "LICENSE" for further details.

#FIXME: Consider moving this submodule into the channel-specific
#"betse.science.tissue.channels" subpackage, for disambiguity.

import numpy as np

# calcium (and voltage) gated K+ channel-------------------------------------------------------------------------------
def cagPotassium(dyna,sim,cells,p):
    """
    Model of a high-conductance calcium activated potassium channel (SK channel), obtained from
    the allosteric model of Cox DH, Cui J, Aldrich RW. J Gen Physiology. 1997. 110: 257-281.
    """

    # get data on cytosolic calcium levels in target cells:

    ca = sim.cc_cells[sim.iCa][cells.mem_to_cells][dyna.targets_cagK]

    # calculate different terms:
    t1 = 1 + ((ca*1e6)/10.22)
    t2 = 1 + ((ca*1e6)/0.89)

    # calculate probability of channel being open or closed:
    P = 1/(1+((t1/t2)**4)*6182*np.exp(-(1.64*p.F*sim.vm)/(p.R*sim.T)))

    # print(P.min(), P.mean(), P.max())

    # # ensure proper probability behaviour:
    # inds_P_over = (P > 1.0).nonzero()
    # P[inds_P_over] = 1.0
    #
    # inds_P_under = (P < 0.0).nonzero()
    # P[inds_P_under] = 0.0

    # calculate conductance of this potassium channel:
    # print(P.mean())

    # delta_Q = - (dyna.maxDmKcag * P * (sim.vm - self.vrev))
    sim.Dm_cag[sim.iK] = dyna.maxDmKcag*P



#!/usr/bin/env python3
# --------------------( LICENSE                           )--------------------
# Copyright 2014-2025 by Alexis Pietak & Cecil Curry.
# See "LICENSE" for further details.

'''
**Simulation export subconfiguration** (i.e., YAML-backed list item configuring
the exportation of one or more external files from the ``plot`` subcommand when
passed the simulation configuration file containing this item) superclasses.
'''

# ....................{ IMPORTS                           }....................
from betse.lib.yaml.abc.yamllistabc import YamlListItemABC
from betse.lib.yaml.abc.yamlmixin import (
    YamlBooledMixin, YamlNamedMixin, YamlTypedMixin)
# from betse.util.type.types import type_check

# ....................{ SUPERCLASSES                      }....................
class SimConfExportABC(
    YamlBooledMixin,
    YamlNamedMixin,
    YamlTypedMixin,
    YamlListItemABC,
):
    '''
    Abstract base class of all **simulation export subconfiguration** (i.e.,
    YAML-backed list item configuring the exportation of one or more external
    files from the ``plot`` subcommand when passed the simulation configuration
    file containing this item) subclasses.
    '''

    # That's all she wrote.
    pass



#!/usr/bin/env python3
# --------------------( LICENSE                           )--------------------
# Copyright 2014-2025 by Alexis Pietak & Cecil Curry.
# See "LICENSE" for further details.

'''
Low-level **call stack** (i.e., list of objects synopsizing all function and
method calls leading to the current function or method call) functionality.

Caveats
----------
Call stack inspection typically assumes the current Python interpreter to be
**CPython** (i.e., the official Python interpreter) rather than a third-party
alternative (e.g., PyPy, Pyston), implying that functions defined by this
module should typically *never* be called by production code. Consider calling
these functions only where needed.

See Also
----------
:func:`betse.util.io.stderr.output_traceback`
    Printing the current call stack to standard error.
:func:`betse.util.io.stdout.output_traceback`
    Printing the current call stack to standard output.
'''

# ....................{ IMPORTS                           }....................
import inspect, traceback
from betse.exceptions import BetseCallableException
from betse.util.type.types import type_check, CallableTypes

# ....................{ GETTERS                           }....................
@type_check
def get_traceback(call_stack_index_last: int = -1) -> str:
    '''
    Human-readable traceback of the current call stack, ignoring exceptions and
    excluding _all_ stack frames inclusively following the stack frame with the
    passed index.

    Parameters
    ----------
    call_stack_index_last : int
        0-based index of the first call stack frame to be ignored. Defaults to
        -1, in which case only the call to this function is ignored.
    '''

    # Call stack excluding the call to the current function.
    call_stack = traceback.extract_stack()[:call_stack_index_last]

    # Human-readable header preceding this traceback.
    traceback_header = 'Traceback (most recent call last):\n'

    # Return this traceback.
    return traceback_header + ''.join(traceback.format_list(call_stack))

# ....................{ GETTERS ~ basename                }....................
@type_check
def get_caller_basename(call_stack_index: int = 2) -> str:
    '''
    **Basename** (i.e., unqualified name *not* preceded by the ``.``-delimited
    name of the parent module or class) of the callable with the passed index
    on the call stack if any *or* raise an exception otherwise.

    Parameters
    ----------
    call_stack_index : optional[int]
        0-based index of the call stack frame to be inspected. Equivalently,
        the 1-based number of leading (most recent) stack frames to be ignored
        including the call to this function. Defaults to an index inspecting
        the **caller's caller** (i.e., the function or method calling the
        function or method calling this function).

    Returns
    ----------
    str
        Basename of this callable, equivalent to the callable after ignoring
        the passed number of leading stack frames on the call stack. As
        example, if passed 2, this is the basename of the second callable on
        the call stack corresponding to that of the caller's caller.

    Raises
    ----------
    BetseCallableException
        If this index exceeds the length of the call stack.
    '''

    # Call stack, including the call to this function.
    call_stack = inspect.stack()

    # Caller stack frame and frame metadata, nullified to avoid exceptions on
    # deleting these locals in the "finally" block below.
    caller_frame = None
    caller_frame_metadata = None

    # Attempt to...
    try:
        # Last valid index into the call stack.
        call_stack_index_max = len(call_stack) - 1

        # If no such callable exists, raise an exception.
        if call_stack_index > call_stack_index_max:
            raise BetseCallableException(
                'Call stack frame {} not found '
                '(i.e., not in range [0, {}]).'.format(
                    call_stack_index, call_stack_index_max))

        # Caller stack frame metadata as an instance of the 5-tuple
        # "(frame, filename, lineno, function, code_context, index)".
        caller_frame_metadata = call_stack[call_stack_index]

        # Caller stack frame as a "FrameType" instance with these fields:
        #
        # * "f_back", next outer frame object (this frame's caller).
        # * "f_builtins", built-in namespace seen by this frame.
        # * "f_code", code object being executed in this frame.
        # * "f_globals", global namespace seen by this frame.
        # * "f_lasti", index of last attempted instruction in bytecode.
        # * "f_lineno", current line number in Python source code.
        # * "f_locals", local namespace seen by this frame.
        # * "f_trace", tracing function for this frame or "None".
        caller_frame = caller_frame_metadata[0]

        # Return the basename of the callable associated with this frame's code
        # object. Lo, and the "inspect" module's API doth mightily sucketh!
        return caller_frame.f_code.co_name
        # return caller_frame.f_func.__qualname__
    # For safety, explicitly release *ALL* call stack frames obtained above.
    # Failing to do so invites memory leaks due to circular references. See:
    #     https://docs.python.org/3/library/inspect.html#the-interpreter-stack
    finally:
        del call_stack, caller_frame_metadata, caller_frame


@type_check
def get_caller_basename_first_matching(predicate: CallableTypes) -> str:
    '''
    Basename of the first callable whose basename matches the passed predicate
    in the call stack if any *or* raise an exception otherwise.

    Parameters
    ----------
    predicate : CallableTypes
        Callable iteratively passed the basename of each callable on the call
        stack (starting at the call to this function and iterating up the call
        stack), returning ``True`` only if that basename matches this
        predicate.

    Returns
    ----------
    str
        Basename of the first callable matching this predicate.

    Raises
    ----------
    BetseCallableException
        If no callable on the call stack matches this predicate.

    See Also
    ----------
    :func:`get_caller_basename`
        Further details.
    '''

    # Call stack, including the call to this function.
    call_stack = inspect.stack()

    # Caller stack frame and frame metadata, nullified to avoid exceptions on
    # deleting these locals in the "finally" block below.
    caller_frame = None
    caller_frame_metadata = None

    # Attempt to...
    try:
        # For each stack frame on the call stack as an instance of the 5-tuple
        # "(frame, filename, lineno, function, code_context, index)"...
        for caller_frame_metadata in call_stack:
            # Caller stack frame as a "FrameType" instance with these fields:
            #
            # * "f_back", next outer frame object (this frame's caller).
            # * "f_builtins", built-in namespace seen by this frame.
            # * "f_code", code object being executed in this frame.
            # * "f_globals", global namespace seen by this frame.
            # * "f_lasti", index of last attempted instruction in bytecode.
            # * "f_lineno", current line number in Python source code.
            # * "f_locals", local namespace seen by this frame.
            # * "f_trace", tracing function for this frame or "None".
            caller_frame = caller_frame_metadata[0]

            # Caller stack frame code object if any or None otherwise.
            caller_frame_code = getattr(caller_frame, 'f_code', None)

            # If this frame is associated with a code object...
            if caller_frame_code is not None:
                # Basename of the callable associated with this frame's code
                # object if any or None otherwise.
                caller_basename = getattr(caller_frame_code, 'co_name', None)

                # If this basename both exists and matches this predicate,
                # return this basename.
                if caller_basename is not None and predicate(caller_basename):
                    return caller_basename

        # Else, no callable on the call stack matches this predicate. In this
        # case, raise an exception.
        raise BetseCallableException(
            'Caller basename matching predicate {!r} not found.'.format(
                predicate))
    # For safety, explicitly release *ALL* call stack frames obtained above.
    # Failing to do so invites memory leaks due to circular references. See:
    #     https://docs.python.org/3/library/inspect.html#the-interpreter-stack
    finally:
        del call_stack, caller_frame_metadata, caller_frame

# ....................{ GETTERS ~ module                  }....................
@type_check
def get_caller_module_name(call_stack_index: int = 2) -> str:
    '''
    Fully-qualified name of the parent module of the callable with the passed
    index on the call stack if any *or* raise an exception otherwise.

    Parameters
    ----------
    call_stack_index : optional[int]
        0-based index of the call stack frame to be inspected. Equivalently,
        the 1-based number of leading (most recent) stack frames to be ignored
        including the call to this function. Defaults to an index inspecting
        the **caller's caller** (i.e., the function or method calling the
        function or method calling this function).

    Returns
    ----------
    str
        Fully-qualified name of the parent module of this callable, equivalent
        to the callable after ignoring the passed number of leading stack
        frames on the call stack. As example, if passed 2, this is the name of
        the parent module of the second callable on the call stack
        corresponding to that of the caller's caller.

    Raises
    ----------
    BetseCallableException
        If this index exceeds the length of the call stack.

    See Also
    ----------
    https://www.calazan.com/how-to-retrieve-the-name-of-the-calling-module-in-python
        Article strongly inspiring this implementation.
    '''

    # Call stack, including the call to this function.
    call_stack = inspect.stack()

    # Caller stack frame metadata, nullified to avoid exceptions on deleting
    # these locals in the "finally" block below.
    caller_frame_metadata = None

    # Attempt to...
    try:
        # Last valid index into the call stack.
        call_stack_index_max = len(call_stack) - 1

        # If no such callable exists, raise an exception.
        if call_stack_index > call_stack_index_max:
            raise BetseCallableException(
                'Call stack frame {} not found '
                '(i.e., not in range [0, {}]).'.format(
                    call_stack_index, call_stack_index_max))

        # Caller stack frame metadata as a 5-tuple
        # "(frame, filename, lineno, function, code_context, index)".
        caller_frame_metadata = call_stack[call_stack_index]
        # print(f'caller_frame_metadata={caller_frame_metadata.frame.f_globals}')

        # Return the fully-qualified name of the parent module of the callable
        # associated with this frame.
        return caller_frame_metadata.frame.f_globals['__name__']
        # return inspect.getmodulename(caller_frame_metadata[1])
    # For safety, explicitly release *ALL* call stack frames obtained above.
    # Failing to do so invites memory leaks due to circular references. See:
    #     https://docs.python.org/3/library/inspect.html#the-interpreter-stack
    finally:
        del call_stack, caller_frame_metadata



#!/usr/bin/env python3
# --------------------( LICENSE                           )--------------------
# Copyright 2014-2025 by Alexis Pietak & Cecil Curry.
# See "LICENSE" for further details.

'''
Low-level **callback** (i.e., caller-defined object external to this
application whose methods are called by other methods internal to this
application as a means of facilitating inter-package communication)
functionality.
'''

# ....................{ IMPORTS                           }....................
from betse.exceptions import BetseCallbackException
from betse.util.io.log import logs
from betse.util.type.types import type_check, StrOrNoneTypes

# ....................{ SUPERCLASSES                      }....................
class CallbacksBC(object):
    '''
    Concrete base class of all **callbacks** (i.e., caller-defined object
    external to this application whose methods are called by other methods
    internal to this application as a means of facilitating inter-package
    communication) subclasses.

    Each of the public methods defined by this superclass is a callback
    intended to be called *only* by the **source callable** (i.e., function or
    method calling each such callback), which receives this object from the
    **sink callable** (i.e., function or method instantiating this object and
    passing this object to the source callable).

    Attributes
    ----------
    _progress_min : int
        Minimum value passed to the most recent call to the
        :meth:`progress_ranged` callback by the source callable. While commonly
        0, the sink callable should *not* assume this to be the case.
    _progress_max : int
        Maximum value passed to the most recent call to the
        :meth:`progress_ranged` callback by the source callable.
    _progress_next : int
        Next value to be implicitly passed to the next call to the
        :meth:`progressed` callback by the :meth:`progressed_next` callback.
    '''

    # ..................{ INITIALIZERS                      }..................
    def __init__(self) -> None:
        '''
        Initialize this callbacks object.
        '''

        # Nullify all instance variables for safety.
        self._clear_progress()

    # ..................{ CALLBACKS ~ progress              }..................
    @type_check
    def progress_ranged(
        self, progress_max: int, progress_min: int = 0) -> None:
        '''
        Callback passed the range of progress values subsequently passed to the
        :meth:`progressed` callback by the source callable.

        Design
        ----------
        The source callable should explicitly call the following callbacks
        *before* performing significant work (in order):

        #. This callback exactly once. Doing so enables the sink callable to
           configure external objects (e.g., progress bars) requiring this
           range.
        #. Either:

           * The :meth:`progressed_next` callback. (Recommended.)
           * The :meth:`progressed` callback with the minimum progress value
             passed to this callback. Since doing so is equivalent to calling
             the :meth:`progressed_next` callback, calling the latter comes
             recommended.

        In either case, the call to the second such callback initializes the
        same external objects (e.g., progress bars) to the minimum progress of
        this range. By design, this :meth:`progress_ranged` callback does *not*
        implicitly call either the :meth:`progressed` or
        :meth:`progressed_next` callbacks on your behalf. Doing so would invite
        chicken-and-egg subtleties with subclasses overriding this method to
        configure external objects.

        Subclasses are advised to reimplement this callback and call this
        superclass implementation in their own reimplementation.

        Parameters
        ----------
        progress_min : optional[int]
            Minimum value subsequently passed to the :meth:`progressed`
            callback by the source callable. Defaults to 0, ensuring the
            ``progress_max`` parameter to be exactly equal to the number of
            times that the :meth:`progressed_next` callback is called by the
            source callable. Nonetheless, the sink callable should *not* assume
            this value to always be 0.
        progress_max : int
            Maximum value subsequently passed to the :meth:`progressed`
            callback by the source callable. If the ``progress_min`` parameter
            is 0, this is equivalent to the number of times that the
            :meth:`progressed_next` callback is called by the source callable.

        Raises
        ----------
        :class:`BetseCallbackException`
            If either:

            * The source callable is still performing work (i.e., the
              :meth:`progressed` callback has yet to be called with a progress
              value equal to ``progress_max``).
            * ``progress_min`` is greater than or equal to ``progress_max``.
        '''

        # If this range is invalid, raise an exception.
        if progress_min > progress_max:
            raise BetseCallbackException(
                'Minimum progress {} exceeds maximum progress {}.'.format(
                    progress_min, progress_max))
        elif progress_min == progress_max:
            raise BetseCallbackException(
                'Minimum progress {} equals maximum progress {}.'.format(
                    progress_min, progress_max))

        # If the source callable is still performing work, raise an exception.
        self._die_if_progressing()

        # Classify all passed parameters for subsequent reference by the
        # progressed() callback.
        self._progress_min = progress_min
        self._progress_max = progress_max

        # Value to be subsequently passed to the progressed() callback by the
        # first call to the progressed_next() callback. See the
        # progressed_next() callback for a justification of the "+ 1" here.
        self._progress_next = progress_min + 1
        # logs.log_debug('next progress: %d', self._progress_next)


    @type_check
    def progress_stated(self, status: str) -> None:
        '''
        Callback passed the current state of progress for work performed by the
        source callable as an arbitrary human-readable string.

        The superclass implementation of this callback logs this string at the
        :attr:`LogLevel.INFO` level, under the generally safe assumption that
        this string is of reasonable relevance to developers *and* users alike.

        Design
        ----------
        The source callable should call this callback repeatedly while
        performing any significant work -- ideally either immediately before or
        after calling the :meth:`progressed`, :meth:`progressed_last`, or
        :meth:`progressed_next` callbacks. Doing so enables the sink callable
        to incrementally update external objects (e.g., status bars) in
        synchronicity with numeric-based progress updates.

        The source callable may technically call this callback at any arbitrary
        time. Unlike the :meth:`progress_ranged` callback and the
        aforementioned callbacks, this callback is purely subjective and hence
        *not* constrained by hard contractual obligations. For usability by the
        sink callable, however, this callable should typically *only* be called
        at approximately the same time as the aforementioned callbacks are
        called by the source callable; doing so increases the likelihood that
        the sink callable is prepared to handle this callback properly.

        Subclasses are advised to reimplement this callback and call this
        superclass implementation in their own reimplementation.

        Parameters
        ----------
        status : str
            Current state of progress for work performed by the source
            callable as an arbitrary human-readable string. To enable the sink
            callable to safely embed this string in size-constrained regions,
            this string should be terse and ideally at most a single line.

        See Also
        ----------
        :meth:`progressed_next`
        :meth:`progressed_last`
            Higher-level callbacks wrapping this lower-level callback in a
            safer manner complying with the above recommendations -- notably,
            that this callback be called in unison with the :meth:`progressed`
            callback by the source callable.
        '''

        # Log this string as an informational message.
        logs.log_info(status)

    # ..................{ CALLBACKS ~ progressed            }..................
    @type_check
    def progressed(self, progress: int) -> None:
        '''
        Callback passed the current state of progress for work performed by the
        source callable.

        If this is the last possible progress value (i.e., if the passed
        ``progress`` parameter is equal to the ``progress_max`` parameter
        previously passed to the :meth:`progress_ranged` callback), this method
        additionally records the source callable to no longer be performing
        work by calling the :meth:`_clear_progress` method. This preserves
        contractual guarantees, including the guarantee that the next call to
        either this or the :meth:`progressed_next` callback will raise an
        exception.

        Design
        ----------
        The source callable should call this callback repeatedly while
        performing any significant work. Doing so enables the sink callable to
        incrementally update external objects (e.g., progress bars).

        Subclasses are advised to reimplement this callback and call this
        superclass implementation in their own reimplementation.

        Parameters
        ----------
        progress : int
            Current state of progress for work performed by the source
            callable. If this integer is *not* in the range previously passed
            to the :meth:`progress_ranged` callback, an exception is raised. If
            this is the first call to this callback by the source callable,
            this integer should ideally be one larger than the minimum progress
            value previously passed to the :meth:`progress_ranged` callback.
            See the :meth:`progressed_next` callback for further discussion.

        Raises
        ----------
        :class:`BetseCallbackException`
            If either:

            * The :meth:`progress_ranged` callback has yet to be called.
            * This progress is *not* in the range previously passed to the
              :meth:`progress_ranged` callback by this source callable.

        See Also
        ----------
        :meth:`progressed_next`
        :meth:`progressed_last`
            Higher-level callbacks automatically passing the next progress
            value according to the the range previously passed to the
            :meth:`progress_ranged` callback.
        '''

        # If progress_ranged() has yet to be called, raise an exception.
        self._die_unless_progressing()

        # If this progress is *NOT* in the range previously passed to
        # progress_ranged(), raise an exception.
        if not (self._progress_min <= progress <= self._progress_max):
            raise BetseCallbackException(
                'Progress {} not in range [{}, {}].'.format(
                    progress, self._progress_min, self._progress_max))

        # If this is the last possible progress value, record the source
        # callable to no longer be performing work.
        if progress == self._progress_max:
            self._clear_progress()

    # ..................{ CALLBACKS ~ progress : next       }..................
    def progressed_next(self, status: StrOrNoneTypes = None) -> None:
        '''
        Higher-level callback wrapping the lower-level :meth:`progressed` and
        :meth:`progress_stated` callbacks by automatically passing the next
        progress value to tha former callback according to the range previously
        passed to the :meth:`progress_ranged` callback *and* passing the passed
        human-readable string if any to the :meth:`progress_stated` callback.

        Note that callers recalling this callback a predetermined number of
        times are advised to replace the last call to this callback
        with a call to the higher-level :meth:`progressed_last` callback; doing
        so guarantees that the last progress value passed to the sink callable
        is the maximum progress value previously passed to the
        :meth:`progress_ranged` callback, preserving an essential contractual
        obligation of this API.

        Guarantees
        ----------
        The first call to this callback is guaranteed to emit one greater than
        the minimum progress value previously passed to the
        :meth:`progress_ranged` callback. While arguably unintuitive, this
        behaviour corresponds with expectations in the sink callable.

        Assuming that the minimum progress value previously passed to the
        :meth:`progress_ranged` callback was 0 as is the common case, this
        behaviour has the additional advantage of guaranteeing that the maximum
        progress value previously passed to that callback is exactly equal to
        the number of times that this :meth:`progressed_next` callback is
        called by the source callable.

        Specifically, since the sink callable typically handles the
        :meth:`progress_ranged` callback by configuring a progress bar to the
        passed range, the sink callable is assumed to have already "consumed"
        the first progress value for all intents and purposes; instructing the
        sink callable to "reconsume" the same progress value would effectively
        hide the actual first iteration of work performed by the source
        callable from the end user perspective.

        Design
        ----------
        For generality, the sink callable should typically *only* handle the
        lower-level :meth:`progressed` and :meth:`progress_stated` callbacks.
        This callback is merely syntactic sugar conveniently simplifying the
        implementation of the source callback calling this callback.

        Parameters
        ----------
        status : StrOrNoneTypes
            Current state of progress for work performed by the source
            callable as an arbitrary human-readable string. To enable the sink
            callable to safely embed this string in size-constrained regions,
            this string should be terse and ideally at most a single line.
            Defaults to ``None``, in which case the :meth:`progress_stated`
            callback is *not* called by this higher-level callback.

        Raises
        ----------
        :class:`BetseCallbackException`
            If the :meth:`progress_ranged` callback has yet to be called.

        See Also
        ----------
        :meth:`progressed`
        :meth:`progress_stated`
            Lower-level callbacks wrapped by this higher-level callback.
        '''

        # If progress_ranged() has yet to be called, raise an exception.
        self._die_unless_progressing()

        # Notify the sink callback of the current state of progress.
        self.progressed(progress=self._progress_next)

        # If the source callback has work remaining to perform, increment this
        # state *AFTER* notifying this callback.
        if self._progress_next is not None:
            self._progress_next += 1

        # If passed status, notify the sink callback of this status.
        if status is not None:
            self.progress_stated(status)


    def progressed_last(self, *args, **kwargs) -> None:
        '''
        Higher-level callback wrapping the lower-level :meth:`progressed_next`
        callback by internally calling that callback and raising an exception
        if the current progress value is *not* the maximum progress value
        previously passed to the :meth:`progress_ranged` callback.

        This callback is intended to be called by callers re-calling the
        :meth:`progressed_next` callback a predetermined number of times as a
        safer replacement for the last such call, ensuring that the last
        progress value passed to the sink callable is indeed the maximum.

        Parameters
        ----------
        All positional and keyword arguments are passed as is to the
        lower-level :meth:`progressed_next` callback wrapped by this callback.

        Raises
        ----------
        :class:`BetseCallbackException`
            If either:

            * The :meth:`progress_ranged` callback has yet to be called.
            * The current progress value is *not* the maximum progress value.

        See Also
        ----------
        :meth:`progressed_next`
            Lower-level callback wrapped by this higher-level callback.
        '''

        # Notify the sink callback of the current state of progress.
        self.progressed_next(*args, **kwargs)

        # If the source callable is still performing work, raise an exception.
        self._die_if_progressing()

    # ..................{ EXCEPTIONS                        }..................
    def _die_if_progressing(self) -> None:
        '''
        Raise an exception if the source callable is still performing work.

        Raises
        ----------
        :class:`BetseCallbackException`
            If the source callable is currently performing work.

        See Also
        ----------
        :meth:`_is_progressing`
            Further details.
        '''

        if self._is_progressing:
            raise BetseCallbackException(
                'progress_ranged() callback called before '
                'progressed() callback called with maximum progress value.')


    def _die_unless_progressing(self) -> None:
        '''
        Raise an exception unless the source callable is still performing work.

        Raises
        ----------
        :class:`BetseCallbackException`
            If the source callable is *not* currently performing work.

        See Also
        ----------
        :meth:`_is_progressing`
            Further details.
        '''

        if not self._is_progressing:
            raise BetseCallbackException(
                'progressed()-style callback called before '
                'progress_ranged() callback '
                '(e.g., due to previously passing a maximum progress value to '
                'progressed(), resetting progress state and requiring calling '
                'progress_ranged() again).'
            )

    # ..................{ TESTERS                           }..................
    @property
    def _is_progressing(self) -> bool:
        '''
        ``True`` only if the source callable is still performing work.

        Equivalently, this method returns ``False`` only:

        * If the source callable is *not* currently performing work.
        * If the state of work performed by this callable has been cleared
          (i.e., to ``None`` values) rather than set to non-``None`` integers.
        * Unless the :meth:`progress_ranged` callback has been called more
          recently than the :meth:`_clear_progress` method.
        '''

        return (
            self._progress_min  is not None and
            self._progress_max  is not None and
            self._progress_next is not None
        )

    # ..................{ CLEARERS                          }..................
    def _clear_progress(self) -> None:
        '''
        Record the source callable to no longer be performing work by clearing
        (i.e., resetting) the state of work performed by this callable to
        ``None`` values.
        '''

        # Nullify all instance variables for safety.
        self._progress_min = None
        self._progress_max = None
        self._progress_next = None



#!/usr/bin/env python3
# --------------------( LICENSE                           )--------------------
# Copyright 2014-2025 by Alexis Pietak & Cecil Curry.
# See "LICENSE" for further details.

'''
**Microtubules** (i.e., vectors aligned in the direction of endogenous current
flux through cells) functionality.
'''

# ....................{ IMPORTS                           }....................
import numpy as np
from betse.lib.numpy import nparray
from betse.science.math import modulate
from betse.util.io.log import logs

# ....................{ CLASSES                           }....................
class Mtubes(object):
    '''
    High-level **microtubules modeller** (i.e., object encapsulating all
    cellular microtubules for the cell cluster) simulated at the current
    simulation time step.

    Attributes (Scalars)
    ----------
    Dr : float
        Microtubule rotational diffusion constant.
    charge_mtube : float
        Charge in Coulombs (C) on each microtubule.
    drag_r : float
        Radial drag force on the microtubule from the Stokes-Einstein relation
        and "Rotational Diffusion" theory.
    p_ind : float
        Inducible polarizability of tubulin in units of [C m^2/V].
    p_mtube : float
        Microtubule dipole moment in units of [C m]. This moment is assumed to
        be the sum of tubulin dimer dipole moments, which are 1740 D each.
    rmt : float
        Radius in meters (m) of each microtubule.
    tubulin_N : float
        Total number of tubulin molecules per microtubule.

    Attributes (Arrays)
    ----------
    mt_density : {ndarray, float}
        Inhomogeneous density. If :attr:`mt_space_density` is ``None``, this is
        a Numpy array; else, this is a float.

    Attributes (Arrays: Vector)
    ----------
    mtubes_x : ndarray
        One-dimensional Numpy array indexing each cell membrane such that each
        item is the normalized X component of the microtubule unit vector
        spatially situated at the midpoint of that membrane for this time step.
    mtubes_y : ndarray
        One-dimensional Numpy array indexing each cell membrane such that each
        item is the normalized Y component of the microtubule unit vector
        spatially situated at the midpoint of that membrane for this time step.
    mtubes_xo : ndarray
        One-dimensional Numpy array indexing each cell membrane such that each
        item is the non-normalized X component of the microtubule vector
        spatially situated at the midpoint of that membrane for this time step.
    mtubes_yo : ndarray
        One-dimensional Numpy array indexing each cell membrane such that each
        item is the non-normalized Y component of the microtubule vector
        spatially situated at the midpoint of that membrane for this time step.
    '''

    # ..................{ INITIALIZORS                      }..................
    def __init__(self, sim, cells, p) -> None:

        # Initialize all core scalar parameters pertaining to microtubules.
        self._init_scalars(cells, p)

        # initial angle of microtubules:
        self.mt_theta = np.arctan2(
            cells.mem_vects_flat[:,3], cells.mem_vects_flat[:,2])

        # normalized microtubule vectors from the cell centre point:
        self.mtubes_x = cells.mem_vects_flat[:,2]*self.mt_density
        self.mtubes_y = cells.mem_vects_flat[:,3]*self.mt_density

        # microtubule density function initialized:
        mtdx = np.dot(cells.M_sum_mems, self.mtubes_x*cells.mem_sa) / cells.cell_sa
        mtdy = np.dot(cells.M_sum_mems, self.mtubes_y*cells.mem_sa) / cells.cell_sa

        self.mtdf = ((mtdx[cells.mem_to_cells]*cells.mem_vects_flat[:,2] +
                                         mtdy[cells.mem_to_cells]*cells.mem_vects_flat[:,3]))

        # Initialize the modulator as a chemical factor altering microtubule
        # dynamics.
        self.modulator = np.ones(sim.mdl)


    def _init_scalars(self, cells, p) -> None:
        '''
        Initialize all core scalar parameters pertaining to microtubules.
        '''

        # Radius of a microtubule.
        self.rmt = p.mt_radius

        self.L = self.rmt * (cells.R_rads/cells.R_rads.mean())

        # Total number of tubulin molecules per microtubule.
        self.tubulin_N = (222/1.0e-6)*self.L

        # Charge on the microtubule in Coulombs.
        self.charge_mtube = (p.length_charge*p.q*self.L)/1.0e-6

        # Microtubule dipole moment [C m] (assumed to be sum of tubulin dimer
        # dipole moments, which are 1740 D each).
        self.p_mtube = self.tubulin_N*p.tubulin_dipole*3.33e-30

        # Inducible polarizability of tubulin C m2/V.
        self.p_ind = self.tubulin_N*p.tubulin_polar*1.0e-40

        # Calculate microtubule drag co-efficients.
        self._init_scalars_drag(cells)

        # Radial drag force on the microtubule from Stokes-Einstein relation
        # "Rotational Diffusion" theory.
        self.drag_r = self.C_rad*p.cytoplasm_viscocity*(self.L**3)

        # Microtubule rotational diffusion constant.
        self.Dr = ((p.kb*p.T)/self.drag_r)

        # Inhomogeneous density.
        if p.mt_space_density is not None:
            self.mt_density, _ = modulate.gradient_bitmap(
                cells.mem_i, cells, p, bitmap_filename=p.mt_space_density)
        else:
            self.mt_density = 1.0


    def _init_scalars_drag(self, cells) -> None:
        '''
        Calculate microtubule drag co-efficients from the drag force on
        cylinders given by Broersma, et al. and Hunt, et al. (1994).
        '''

        v = 1 / (np.log(self.L/self.rmt))

        g_para = -0.114 - 0.15 * v - 13.5 * (v ** 2) + 37 * (v ** 3) - 22 * (v ** 4)
        g_perp = 0.866 - 0.15 * v - 8.1 * (v ** 2) + 18 * (v ** 3) - 9 * (v ** 4)
        g_rad = -0.446 - 0.2 * v - 16 * (v ** 2) + 63 * (v ** 3) - 62 * (v ** 4)

        self.C_para = (2 * np.pi) / (np.log(self.L / (2 * self.rmt)) + g_para)
        self.C_perp = (4 * np.pi) / (np.log(self.L / (2 * self.rmt)) + g_perp)
        self.C_rad = ((1 / 3) * np.pi) / (np.log(self.L / (2 * self.rmt)) + g_rad)


    def reinit(self, cells, p) -> None:
        '''
        Reinitialize key microtubule parameters that may have changed in the
        simulation configuration file since the prior initialization (if any).
        '''

        # Initialize all core scalar parameters pertaining to microtubules.
        self._init_scalars(cells, p)

        # smoothing weights for membrane and central values:
        nfrac = p.smooth_cells
        self.smooth_weight_mem = ((nfrac*cells.num_mems[cells.mem_to_cells] -1)/(nfrac*cells.num_mems[cells.mem_to_cells]))
        self.smooth_weight_o = 1/(nfrac*cells.num_mems[cells.mem_to_cells])

        # amt = self.L/(2*self.rmt)
        # self.drag_r = (np.pi*p.cytoplasm_viscocity*(self.L**3))/(np.log(amt - 0.662 + (0.917/amt) - (0.05/amt**2)))

        if p.mtube_init_x is not None:
            # Initialize microtubules with mini-simulation to define initial
            # state.
            self.Phi_orient_x, _ = modulate.gradient_bitmap(
                cells.mem_i, cells, p, bitmap_filename=p.mtube_init_x)
        else:
            # self.Phi_orient_x = np.zeros(len(cells.mem_i))
            self.Phi_orient_x = cells.mem_vects_flat[:, 2]

        if p.mtube_init_y is not None:
            self.Phi_orient_y, _ = modulate.gradient_bitmap(
                cells.mem_i, cells, p, bitmap_filename=p.mtube_init_y)
        else:
            self.Phi_orient_y = cells.mem_vects_flat[:, 3]
            # self.Phi_orient_y = np.zeros(len(cells.mem_i))
        # if self.Phi_orient_x != cells.mem_vects_flat[:, 2] or self.Phi_orient_y != cells.mem_vects_flat[:,3]:

        # if there's a user-requested gradient for initial condition alignment, then perform the pre-simulation:
        self.presim(self.Phi_orient_x, self.Phi_orient_y, cells, p)

        self.uxmt, self.uymt = self.mtubes_to_cell(cells, p)

    # ..................{ UPDATERS                          }..................
    def update_mtubes(self, cells, sim, p) -> None:

        # microtubule radial vectors:
        ui = np.cos(self.mt_theta)*self.L
        vi = np.sin(self.mt_theta)*self.L

        # nx = cells.mem_vects_flat[:,2]
        # ny = cells.mem_vects_flat[:,3]

        Ex = sim.E_cell_x[cells.mem_to_cells]
        Ey = sim.E_cell_y[cells.mem_to_cells]

        gEx = (Ex[cells.nn_i] - Ex[cells.mem_i]) / (cells.nn_len)

        gExx = gEx*cells.nn_tx
        gExy = gEx*cells.nn_ty

        gEy = (Ey[cells.nn_i] - Ey[cells.mem_i]) / (cells.nn_len)

        gEyx = gEy*cells.nn_tx
        gEyy = gEy*cells.nn_ty

        q_tube = self.charge_mtube


        if p.tethered_tubule is False:

            torque_tether = np.zeros(sim.mdl)

            # gradient of the field will torque the monopole by applying different forces at ends:
            torque_gradient = (q_tube * (ui) * (gEyx.ravel() * ui + gEyy.ravel() * vi) -
                               q_tube * (vi) * (gExx.ravel() * ui + gExy.ravel() * vi))

            # fiber will also align such that ends are at the same voltage:
            torque_monopole = (q_tube*(ui)*Ex.ravel() + q_tube*(vi)*Ey.ravel())


        # if fiber is tethered, any perpendicular force will represent a torque:
        else:

            torque_tether = (q_tube * ui * Ey.ravel() - q_tube * vi * Ex.ravel())

            # gradient of the field will torque the monopole by applying different forces at ends:
            torque_gradient = np.zeros(sim.mdl)

            # fiber will also align such that ends are at the same voltage:
            torque_monopole = np.zeros(sim.mdl)

            # fiber will also align via its dipole in the electric field:
            # torque_dipole = (self.p_ind * ui_hat * Ey.ravel() - self.p_ind * vi_hat * Ex.ravel())

        flux_theta = (
            + (torque_tether / self.drag_r)
            + (torque_monopole / self.drag_r)
            + (torque_gradient / self.drag_r)
            # + ((p.kb * p.T) / self.drag_r) * (0.5 - np.random.rand(len(self.mt_theta)))
        )

        # update the microtubule coordinates with the new angle:
        if p.dilate_mtube_dt > 0.0:
            # normalized correlation length of the microtubules
            lenmt = np.sqrt(self.uxmt[cells.mem_to_cells] ** 2 + self.uymt[cells.mem_to_cells] ** 2) + p.kb * sim.T

            stdev = np.sqrt(2 * p.dt * p.dilate_mtube_dt * ((p.kb * p.T) / self.drag_r) * lenmt * self.L ** 2)

            noise = np.random.normal(loc=0.0, scale=stdev, size=sim.mdl)

            # update the microtubule angle:
            self.mt_theta = self.mt_theta + flux_theta * p.dt * p.dilate_mtube_dt * self.modulator + noise

            mtubes_xo = np.cos(self.mt_theta)*self.mt_density
            mtubes_yo = np.sin(self.mt_theta)*self.mt_density

            self.mtubes_x, self.mtubes_y = cells.single_cell_div_free(mtubes_xo, mtubes_yo)

            self.uxmt, self.uymt = self.mtubes_to_cell(cells, p)


    def mtubes_to_cell(self, cells, p) -> None:

        # determine the microtubules base electroosmotic velocity:
        uxmto = self.mtubes_x
        uymto = self.mtubes_y

        # averages of mtube field at the cell centres:
        # uxmt = (np.dot(cells.M_sum_mems, uxmto*cells.mem_sa)/cells.cell_sa)
        # uymt = (np.dot(cells.M_sum_mems, uymto*cells.mem_sa)/cells.cell_sa)

        uxmt = (np.dot(cells.M_sum_mems, uxmto)/cells.num_mems)
        uymt = (np.dot(cells.M_sum_mems, uymto)/cells.num_mems)

        # average the mtube field to the centre of pie-shaped midpoints of each individual cell:
        # uxmti = (uxmt[cells.mem_to_cells] + uxmto)/2
        # uymti = (uymt[cells.mem_to_cells] + uymto)/2

        uxmti = uxmt[cells.mem_to_cells]
        uymti = uymt[cells.mem_to_cells]

        # uxmti, uymti = cells.single_cell_div_free(uxmti, uymti)

        # Store the normal component of microtubule alignment field mapped to membranes:
        self.umtn = (uxmti * cells.mem_vects_flat[:, 2] + uymti * cells.mem_vects_flat[:, 3])

        return uxmt, uymt


    def remove_mtubes(
        self, target_inds_mem, target_inds_cell, cells, sim, p) -> None:
        '''
        Remove all microtubules from the cells and cell membranes with the
        passed indices.
        '''

        self.mtubes_x     = np.delete(self.mtubes_x, target_inds_mem)
        self.mtubes_y     = np.delete(self.mtubes_y, target_inds_mem)
        self.mtdf         = np.delete(self.mtdf, target_inds_mem)
        self.mt_theta     = np.delete(self.mt_theta, target_inds_mem)
        self.modulator    = np.delete(self.modulator, target_inds_mem)
        self.L            = np.delete(self.L, target_inds_mem)
        self.charge_mtube = np.delete(self.charge_mtube, target_inds_mem)
        self.tubulin_N    = np.delete(self.tubulin_N, target_inds_mem)
        self.p_ind        = np.delete(self.p_ind, target_inds_mem)
        self.drag_r       = np.delete(self.drag_r, target_inds_mem)
        self.C_perp       = np.delete(self.C_perp, target_inds_mem)
        self.Dr           = np.delete(self.Dr, target_inds_mem)

        # If the "mt_density" instance variable has been conditionally defined
        # by the _init_scalars() method to be a Numpy array rather than a
        # float, remove the required items from this array.
        if nparray.is_array(self.mt_density):
            self.mt_density = np.delete(self.mt_density, target_inds_mem)

        # mtux2 = np.delete(self.uxmt, target_inds_cell)
        # self.uxmt = mtux2*1
        #
        # mtuy2 = np.delete(self.uymt, target_inds_cell)
        # self.uymt = mtuy2*1

        # mtuu2 = np.delete(self.umtn, target_inds_cell)
        # self.umtn = mtuu2*1


    def presim(self, gFxo, gFyo, cells, p) -> None:
        '''
        Initialize microtubule orientation using gradient of a vector field
        '''

        if len(gFxo) == len(cells.cell_i):
            gFxo = gFxo[cells.mem_to_cells]
            gFyo = gFyo[cells.mem_to_cells]

        if p.mtube_init_x is not None and p.mtube_init_y is not None:
            logs.log_info('-------------------------------')
            logs.log_info(
                'Preinitializing microtubule x- and y- coordinates with '
                '%f and %f', p.mtube_init_x,  p.mtube_init_y)
            logs.log_info('-------------------------------')

            # rotate the axis of the model:
            rotangle = (p.mtube_init_rotangle*np.pi)/180.0

            gFx = gFxo*np.cos(rotangle) - gFyo*np.sin(rotangle)
            gFy = gFxo*np.sin(rotangle) + gFyo*np.cos(rotangle)

            # magnitude of the orienting field:
            magF = (np.sqrt(gFx ** 2 + gFy ** 2)) + 1.0e-15

            # set the microtubule vectors with the field values:
            mtubes_xo = (gFx/magF) * self.mt_density
            mtubes_yo = (gFy/magF) * self.mt_density

            self.mtubes_x, self.mtubes_y = cells.single_cell_div_free(mtubes_xo, mtubes_yo)

            # initial angle of microtubules:
            self.mt_theta = np.arctan2(self.mtubes_y, self.mtubes_x)

            self.uxmt, self.uymt = self.mtubes_to_cell(cells, p)



#!/usr/bin/env python3
# Copyright 2014-2025 by Alexis Pietak & Cecil Curry.
# See "LICENSE" for further details.

'''
Abstract base classes of all vector field subclasses.
'''

# ....................{ IMPORTS                            }....................
import numpy as np
from betse.lib.numpy import nparray
from betse.science.math.vector.veccls import VectorCellsCache
from betse.util.type.decorator.decmemo import property_cached
from betse.util.type.types import type_check, SequenceTypes
from numpy import ndarray

# ....................{ CLASSES                            }....................
class VectorField(object):
    '''
    Vector field whose X and Y components are two-dimensional Numpy arrays
    describing a single modelled variable (e.g., total current density)
    spatially situated along one coordinate system (e.g., cell centres, cell
    membrane vertices) for one or more simulation time steps

    This object provides properties efficiently yielding these X and Y
    components in both normalized and nonnormalized forms as well as the
    magnitudes of these components.

    Attributes
    ----------
    x : ndarray
        Two-dimensional sequence whose:
        * First dimension indexes one or more time steps of this simulation.
        * Second dimension indexes each X component of each vector in this
          vector field for this time step.
    y : ndarray
        Two-dimensional sequence whose:
        * First dimension indexes one or more time steps of this simulation.
        * Second dimension indexes each Y component of each vector in this
          vector field for this time step.
    '''

    # ..................{ INITIALIZERS                       }..................
    @type_check
    def __init__(
        self,

        # Mandatory parameters.
        x: SequenceTypes,
        y: SequenceTypes,
    ) -> None:
        '''
        Initialize this vector field.

        Parameters
        ----------
        x : ndarray
            Two-dimensional sequence whose:
            * First dimension indexes one or more time steps of this simulation.
            * Second dimension indexes each X component of each vector in this
              vector field for this time step.
        y : ndarray
            Two-dimensional sequence whose:
            * First dimension indexes one or more time steps of this simulation.
            * Second dimension indexes each Y component of each vector in this
              vector field for this time step.
        '''

        # Classify the passed sequences as Numpy arrays for efficiency.
        self._x = nparray.from_iterable(x)
        self._y = nparray.from_iterable(y)

    # ..................{ PROPERTIES ~ abstract              }..................
    @property
    def x(self) -> ndarray:
        '''
        Two-dimensional Numpy array whose:

        * First dimension indexes one or more time steps of this simulation.
        * Second dimension indexes each X component of each vector in this
          vector field for this time step.
        '''

        return self._x


    @property
    def y(self) -> ndarray:
        '''
        Two-dimensional Numpy array whose:

        * First dimension indexes one or more time steps of this simulation.
        * Second dimension indexes each Y component of each vector in this
          vector field for this time step.
        '''

        return self._y

    # ..................{ PROPERTIES ~ magnitudes            }..................
    @property_cached
    def magnitudes(self) -> ndarray:
        '''
        Two-dimensional Numpy array whose:

        * First dimension indexes one or more time steps of this simulation.
        * Second dimension indexes each (possibly zero) magnitude of each vector
          in this vector field for this time step.

        This array is created only on the first access of this property.

        Caveats
        ----------
        This array may contain zero values and is thus *not* safe for
        general-purpose use as the divisor of division (e.g., to normalize the X
        and Y components of this field). See :meth:`magnitudes_nonzero` for a
        comparable array guaranteed *not* to contain zero values,
        '''

        #FIXME: For efficiency, refactor this to simply call a builtin numpy
        #norm function or method. (Who knew?)

        # Array of all vector magnitudes computed from the arrays of all
        # vector X and Y components.
        return np.sqrt(self._x**2 + self._y**2)


    @property_cached
    def magnitudes_nonzero(self) -> ndarray:
        '''
        Two-dimensional Numpy array whose:

        * First dimension indexes one or more time steps of this simulation.
        * Second dimension indexes each magnitude of each vector in this vector
          field for this time step. For safety, this magnitude is guaranteed to
          be non-zero, avoiding division-by-zero errors elsewhere in the
          codebase when these magnitudes are subsequently divided by (e.g., to
          normalize the X or Y components of this field). If the unmodified
          value of this magnitude in the :attr:`magnitudes` array is zero, this
          magnitude is replaced with 1.0. This magnitude is indistinguishable
          from magnitudes whose unmodified values are also 1.0 and is thus
          safely usable *only* in contexts where vectors with zero magnitudes
          are ignorable (e.g., as the divisior of division).

        This array is created only on the first access of this property.
        '''

        # Array of all vector magnitudes copied from the original.
        magnitudes_nonzero = self.magnitudes.copy()

        # Substitute all zero magnitudes by 1.0.
        magnitudes_nonzero[magnitudes_nonzero == 0.0] = 1.0

        # Return the resulting array.
        return magnitudes_nonzero

    # ..................{ PROPERTIES ~ unit                  }..................
    @property_cached
    def unit_x(self) -> ndarray:
        '''
        Two-dimensional Numpy array whose:

        * First dimension indexes one or more time steps of this simulation.
        * Second dimension indexes each unitary X component of each unit vector
          in this vector field for this time step, produced by dividing that
          vector's original X component by that vector's original magnitude.

        This array is created only on the first access of this property.
        '''

        return self._x / self.magnitudes_nonzero


    @property_cached
    def unit_y(self) -> ndarray:
        '''
        Two-dimensional Numpy array whose:

        * First dimension indexes one or more time steps of this simulation.
        * Second dimension indexes each unitary Y component of each unit vector
          in this vector field for this time step, produced by dividing that
          vector's original Y component by that vector's original magnitude.

        This array is created only on the first access of this property.
        '''

        return self._y / self.magnitudes_nonzero

# ....................{ CLASSES ~ cache                    }....................
class VectorFieldCellsCache(object):
    '''
    Cell cluster vector field cache, persisting all vector fields whose X and Y
    components are two-dimensional Numpy arrays describing the same underlying
    data spatially situated along different coordinate systems (e.g., cell
    centres, cell membrane vertices) for one or more simulation time steps.

    The input data encapsulated by this cache may be spatially situated at
    either:

    * The centre of each cell in the simulated cluster.
    * The vertex of each cell membrane in the simulated cluster.
    * The midpoint of each cell membrane in the simulated cluster.
    * The centre of each square grid space (in either dimension).

    Each property provided by this vector field (e.g.,
    :meth:`times_cell_centres`) then efficiently interpolates this input data
    from its original coordinate system into the corresponding output data in
    another coordinate system.

    Attributes
    ----------
    _field_args : tuple
        Tuple of all positional arguments to be passed as is to the
        :meth:`VectorField.__init__` method of each :class:`VectorField`
        instance returned by each property of this cache.
    _field_kwargs : dict
        Dictionary of all keyword arguments to be passed as is to the
        :meth:`VectorField.__init__` method of each :class:`VectorField`
        instance returned by each property of this cache.
    _field_x : VectorCellsCache
        Two-dimensional array of the X components of all vectors in this field
        for one or more simulation time steps.
    _field_y : VectorCellsCache
        Two-dimensional array of the Y components of all vectors in this field
        for one or more simulation time steps.
    '''

    # ..................{ INITIALIZERS                       }..................
    @type_check
    def __init__(
        self,
        x: VectorCellsCache,
        y: VectorCellsCache,
        *args, **kwargs
    ) -> None:
        '''
        Initialize this vector field cache.

        Parameters
        ----------
        x : VectorCellsCache
            Two-dimensional array of the X components of all vectors in this
            field for one or more simulation time steps.
        y : VectorCellsCache
            Two-dimensional array of the Y components of all vectors in this
            field for one or more simulation time steps.

        All remaining parameters are passed as is to the
        :meth:`VectorField.__init__` method of each :class:`VectorField`
        instance returned by each property of this cache.
        '''

        # Classify all passed parameters.
        self._field_x = x
        self._field_y = y
        self._field_args = args
        self._field_kwargs = kwargs

    # ..................{ PROPERTIES                         }..................
    # Read-only properties, preventing callers from setting these attributes.

    @property_cached
    def times_cells_centre(self) -> VectorField:
        '''
        Vector field whose X and Y components are two-dimensional Numpy arrays
        of data spatially situated at cell centres for one or more time steps.

        This field is created only on the first access of this property.

        See Also
        ----------
        :meth:`VectorCellsCache.times_cells_centre`
            Details on the structure of these arrays.
        '''

        return VectorField(
            *self._field_args,
            x=self._field_x.times_cells_centre,
            y=self._field_y.times_cells_centre,
            **self._field_kwargs
        )


    @property_cached
    def times_grids_centre(self) -> VectorField:
        '''
        Vector field whose X and Y components are two-dimensional Numpy arrays
        of data spatially situated at grid space centres for one or more
        time steps.

        This field is created only on the first access of this property.

        See Also
        ----------
        :meth:`VectorCellsCache.times_grids_centre`
            Details on the structure of these arrays.
        '''

        return VectorField(
            *self._field_args,
            x=self._field_x.times_grids_centre,
            y=self._field_y.times_grids_centre,
            **self._field_kwargs
        )


    @property_cached
    def times_membranes_midpoint(self) -> VectorField:
        '''
        Vector field whose X and Y components are two-dimensional Numpy arrays
        of data spatially situated at cell membrane midpoints for one or more
        time steps.

        This field is created only on the first access of this property.

        See Also
        ----------
        :meth:`VectorCellsCache.times_membranes_midpoint`
            Details on the structure of these arrays.
        '''

        return VectorField(
            *self._field_args,
            x=self._field_x.times_membranes_midpoint,
            y=self._field_y.times_membranes_midpoint,
            **self._field_kwargs
        )


    @property_cached
    def times_membranes_vertex(self) -> VectorField:
        '''
        Vector field whose X and Y components are two-dimensional Numpy arrays
        of data spatially situated at cell membrane vertices for one or more
        time steps.

        This field is created only on the first access of this property.

        See Also
        ----------
        :meth:`VectorCellsCache.times_membranes_vertex`
            Details on the structure of these arrays.
        '''

        return VectorField(
            *self._field_args,
            x=self._field_x.times_membranes_vertex,
            y=self._field_y.times_membranes_vertex,
            **self._field_kwargs
        )
